﻿;;;-*- Mode: Lisp; Package: "ARISTIDE" -*-
;;;===============================================================================
;;;12/01/09
;;;             AGENT ARISTIDE: DIALOG (file ARISTIDE-dialog.lisp)
;;;             Copyright Barthes@UTC, 2006-2012
;;;
;;;=============================================================================== 

#|
ARISTIDE est l'assistant français de Milton Ramos. Aristide ne parle que français.
Il peut récupérer des adresses email des numéros de téléphone ou utiliser des services web (?)

Ce fichier est sa bibliothèque de dialogues.

Exemples de questions que l'on peut poser à ARISTIDE :
   - quel est le numéro de téléphone de Trigano?
     (what is the telephone number of Trigano?)
   - téléphone dbb?
     (phone dbb?)
   - téléphone du domicile de Moulin?
     (home phone of Moulin?)
   - poste de DICIT?
     (extension of DICIT?)
   - prévenir Gapenne.
     (send message to "gapenne".)
   - ...

Les dialogues suivants sont expérimentaux et plutôt simplistes.
   - MAIN DIALOG
   - ASK
   - EXPLAIN
   - GET ADDRESS
   - GET HOME ADDRESS
   - GET NEWS BY SECTION
   - GET PHONE NUMBER
   - GET LAST NEWS
   - PRINT CONCEPT DOCUMENTATION
   - PRINT HELP
   - PROCESS ASK
   - PROCESS TELL
   - PUBLISH NEWS
   - SET FONT SIZE
   - TELL
   - TRACE
   - WHAT IS
   - WHO IS
   - ELIZA

IMPORTANT :  OMAS Content Language
==================================
Le "content language" d'OMAS est simple: ((<tag> . <val>)+) où <tag> est un mot clé
(:data, :query, :find, :pattern, :language)

:data
-----
La valeur associée est une liste de mots : (:data "où" "habite" "Joseph" "?")

:query
------
La valeur associée est une requête MOSS : (:query "personne" ("nom" :is "Ramos"))

:find
-----
La valeur associée est un concept : (:find . "personne") et les arguments sont dans une
clause :data.

:pattern
--------
Précise la forme escomptée de la réponse
- s'il n'y a pas de pattern : la réponse est une liste de strings (litéraux). Le 
résultat peut être transmis ou facilement imprimé.
- sinon pattern doit être une structure CORREPONDANT À L'ONTOLOGIE DE L'ÉMETTEUR :
("personne" ("nom") ("prénom")("adresse"))
dans ce cas le récepteur doit s'arranger pur associer une string à chaque élément du
pattern, exemple
("personne" ("nom" "Barthès")("prénom" "Jean-Paul")("adresse" "UTC, Compiègne"))
le premier élément donne le type d'individus recherchés.
L'émetteur peut alors recontituer un individu dans sa base de connaissance et le 
conserver s'il le souhaite. S'il souhaite l'imprimer il lui faudar fournir une
fonction spéciale.

:language
---------
Sert à préciser le langage pour le récepteur : (:language . :FR)

2009
0730 ceating file by cloning Albert-dialog
2010
0614 upgrading to v8 omas-agent properties start with ag- 
1120 upgrading the content of the file to adapt to v8.0.8
|#

(in-package :ARISTIDE)

;;; on précise ici que le langage d'ARISTIDE est le français
(eval-when (:compile-toplevel :load-toplevel :execute)
  (setq *language* :fr)
  )

(format t "~&loading ARISTIDE dialogs, package is: ~S" *package*)

;;;================================================================================
;;;                      Debugging functions
;;;================================================================================

;;; prints the content of the conversation object
(defMacro pcs () `(moss::%pep (omas::conversation ARISTIDE::PA_ARISTIDE)))

;;; to return to the ARISTIDE package when in CL=USER
(defMacro cl-user::ar () `(in-package :aristide))

;;; to trace some functions (verbose flag) Defined in moss-macros.lisp
;(defUn v+ ()(setq moss::*verbose* t) "moss::*verbose* set to T")
;(defUn v- ()(setq moss::*verbose* nil) "moss::*verbose* set to NIL")

;;; to see the expansion of the defstate macro
;(defUn d+ ()(setq moss::*debug* t) "moss::*debug* set to T")
;(defUn d- ()(setq moss::*debug* nil) "moss::*debug* set to NIL")

;;; to restart a conversation after a crash
(defUn cc () 
  (declare (special _main-conversation))
  (moss::start-conversation _main-conversation))


;;; liste de mots vides qui pourront être supprimés de certaines phrases

(defParameter *empty-words* '("le" "la" "les" "l" "un" "une" "uns" "unes" "de"
                              "du" "d" "quel" "quels" "quelle" "quelles" "est"
                              "sont"))


;;;================================================================================

;;; here the strategy is to develop a specific dialog for each task. However, we 
;;; could also use a generic dialog consisiting in:
;;;    - getting the list of required parameters
;;;    - for each parameter, trying to extract info from data using predefined
;;;      rules or extraction patterns (Note that this is not much different than
;;;      having a predefined dialog, albeit focused on a single parameter)
;;;    - if some required parameters are missing, ask master about them using 
;;;      predefined questions
;;;    - shipping result to ad hoc staff agent

;;; The generic dialog could be an option offered to the programmer, which requires
;;; a suitable modeling of the tasks.

;;;================================================================================
;;;
;;;                              Service functions
;;;
;;;================================================================================

;;; Should be part of the OMAS services

;;;---------------------------------------------------------------- DISPLAY-ANSWER

(defUn display-answer (args)
  "prints the list of returned strings (no pattern used)."
  (omas::assistant-display-text 
   PA_ARISTIDE
   (format nil "~{~A~^~2%~}~%" args)))

;;;================================================================================
;;;
;;;                              Sub-Conversations
;;;
;;;================================================================================

;;; Sub-conversations are defined by a conversation header, that points to the
;;; entry state.
;;; There are two output states:
;;;   SUCCESS: the result is stored into the FACTS/result area of the conversation
;;;            object
;;;   FAILURE: no result could be achieved
;;; In case of bad error there is a throw to an :error label
;;; data are obtained from the FACTS/input area if the conversation object


;;;================================================================================
;;;
;;;                           ASK CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to ask a remote PA about something.
;;; We must first locate our receiver (presumably the PA of somebody in the coterie
;;; We then want to prepare topic and text, then send the message

(defsubdialog 
    _ask-conversation 
    (:label "Ask conversation")
  (:explanation "We want our PA to ask something to somebody.")
  (:states _ask-entry-state  ; required by defstate            
           _ask-choose-destination 
           _ask-get-data
           _ask-get-destination
           _ask-send-message
           _ask-sorry
           )
  )

;;;================================================================================
;;;                       ASK CONVERSATION SERVICE FUNCTIONS
;;;================================================================================

;;;------------------------------------------------------------ EXTRACT-DESTINATION

(defUn extract-destination (text-list)
  "tries to obtain the key of a PA as a destination for a message.
   first try is for PA name, second for person name, third for
   person first name.
Arguments:
   text-list: list of words
Return:
   a PA keyword a list of persons or nil in case of failure."
  (let (entry-point-list key pa)
    (cond
     ;; if text-list is nil, then error
     ((null text-list)
      (return-from extract-destination nil))
     ;; try to locate the name of an agent (a PA)
     ((setq entry-point-list 
            (moss::find-best-entries-from-property-name-and-class-ref
             text-list 'HAS-AG-NAME "omas-agent" :keep-results t))
      ;; agents have unique names, thus take the first element from the list
      (setq key (car (has-ag-key (cadar entry-point-list))))
      (return-from extract-destination key))
     ;; could not find agent, try to locate a person from name or first name
     ((setq entry-point-list 
            (moss::find-best-entries-from-property-name-and-class-ref
             text-list 'HAS-nom "personne" :keep-results t))
      ;; if a list return the list of person IDs
      (when (cdr entry-point-list)
        (return-from extract-destination (mapcar #'cadr entry-point-list))))
     ;; could not find a person using names, try first names
     ((setq entry-point-list 
            (moss::find-best-entries-from-property-name-and-class-ref
             text-list 'HAS-prénom "personne" :keep-results t))
      (when (cdr entry-point-list)
        (return-from extract-destination (mapcar #'cadr entry-point-list))))
     ;; could not find anything
     (t (return-from extract-destination nil)))
    (format t "~%; ARISTIDE extract-destination /entry-point-list: ~%  ~S"
      entry-point-list)
    ;; here we have a single answer
    (setq pa (car (send (cadar entry-point-list) '=get "assistant personnel")))
    ;; and agent key
    (car (has-ag-key pa))))

#|
(extract-destination nil)
NIL
(extract-destination '("albert"))
:ALBERT
(extract-destination '("ramos"))
:ARISTIDE
(extract-destination '("jean-paul"))
:ALBERT
(extract-destination '("omar"))
:ALFREDO
(extract-destination '("envoyer" "un" "message" "à" "milton"))
:ARISTIDE
(extract-destination '("milton" "ramos"))
:ARISTIDE
|# 
;;;================================================================================
;;;                            ASK CONVERSATION STATES
;;;================================================================================

;;;---------------------------------------------------------- (ASK) ASK-ENTRY-STATE

(defstate 
    _ask-entry-state
    (:label "Ask dialog entry")
  (:entry-state _ask-conversation)
  (:explanation "We try to get destination from the input data.")
  (:answer-analysis)
  )

(defownmethod
    =answer-analysis _ask-entry-state (conversation input)
  "Using list of words to locate receiver of message to send. Must ask user ~
   to confirm findings.
Arguments:
   conversation: conversation object
   input: copy of the FACTS/input area"
  (let* (bindings destination)
    (catch 
     :return
     (cond
      ;; when input is empty then failure
      ((null input)
       (throw :return (list :failure)))
      ;; try to extract destination info from the input text
      ((setq bindings
             (MOSS::PROCESS-PATTERNS 
              input
              '((* "demander" "à" ?Y) 
                (* "demander" * "à" ?Y)
                (* "demande" ?Y) 
                (* "question" "pour" ?Y)
                (* "question" ?Y)
                )))
       (setq input (cdr (assoc '?Y bindings)))
       )
      ;; otherwise do nothing
      )
     
     ;; try to get destination
     (setq destination (extract-destination input))
     
     (cond
      ;; if null, failure
      ((null destination)
       (throw :return (list :transition _ask-sorry)))
      ;; list, must make a choice
      ((listp destination)
       ;; save the result into the FACT area and go ask for choice
       (replace-fact conversation :to-choose destination)
       (throw :return (list :transition _ask-choose-destination)))
      ;; otherwise we got destination, save it
      (t
       (replace-fact conversation :to destination)
       ;; and go get data
       (throw :return (list :transition _ask-get-data))))
     )))

;;;------------------------------------------------------- (ASK) CHOOSE-DESTINATION
;;; fake: currently takes the first element of the list of choices

(defstate 
    _ask-choose-destination
    (:label "Ask choose destination")
  (:explanation "ambiguous destination (list) select one.")
  (:execute
   (let* ((conversation moss::conversation)
          (choices (read-fact conversation :to-choose))
          )
     ;; retrieve the temporary data, take the first one in the list
     (replace-fact conversation :to-choose nil)
     (replace-fact conversation :to (car choices))))
  (:transitions (:always :target _ask-get-data)))

;;;----------------------------------------------------------------- (ASK) GET-DATA

(defstate 
    _ask-get-data
    (:label "set up parameters for continuous input.")
  (:explanation "Let the master type text into the pane until the DONE button is ~
                 pushed. We must disable . and ? terminations.")
  ;; set global flag in the interface window so that we can have . and ? in the text
  (:execute-preconditions
   ;; set up a mark in the input window structure (will be removed automatically)
   (setf (omas::pass-every-char (car (has-moss-input-window moss::conversation))) t)
   ;; include :verbatim flag into the FACT base (?)
   (replace-fact moss::conversation :verbatim t)
   )
  (:question-no-erase 
   "~% ...Tapez ou dictez votre texte puis cliquez sur le bouton Terminer.")
  (:transitions
   (:always :target _ask-send-message))
  )

;;;---------------------------------------------------------- (ASK) GET-DESTINATION

(defstate 
    _ask-get-destination
    (:label "obtain master's destination for message.")
  (:explanation "Get the id of the personal assistant of the receiver.")
  ;; set global so that we can have . and ? in the text
  (:question-no-erase 
   "~% A qui voulez-vous envoyer le message ?") 
  (:answer-analysis)
  )

(defownmethod =answer-analysis _ask-get-destination (conversation input)
  "We process the answer and record it unless we get an abort.
Arguments:
   agent: assistant."
  (let (destination)
    ;; analyze input
    (setq destination (extract-destination input))
    
    (cond
     ((null destination)
      (list :transition _ask-sorry))
     ((listp destination)
      (replace-fact conversation :to-choose destination)
      (list :transition _ask-choose-destination))
     (t
      (replace-fact conversation :to destination)
      ;; we then get out of dialog
      (list :transition _ask-get-data)))))

;;;------------------------------------------------------------- (ASK) SEND-MESSAGE

(defstate 
    _ask-send-message
    (:label "Ask send final message")
  (:explanation "build and send the message.")
  (:send-message-no-wait 
   :to (pop-fact moss::conversation :to)
   :type :request
   :action :ask
   :args (list (list :text (pop-fact moss::conversation :raw-input))))
  (:text "~&...Message envoyé")
  (:transitions (:always :success)) ; return
  )

;;;-------------------------------------------------------------------- (ASK) SORRY

(defstate 
    _ask-sorry
    (:label "Ask failure.")
  (:explanation "We don't have the needed parameters")
  (:text 
   "Désolé, je ne vois pas à qui vous voulez envoyer le message.~%")
  (:reset)
  (:transitions (:failure))
  )

;;;================================================================================
;;;
;;;                          EXPLAIN CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to explain the meaning of a concept.

(defsubdialog 
    _explain-conversation 
    (:label "Explain conversation")
  (:explanation "The data is searched for a concept to be explained.")
  (:states _exp-entry-state  ; required by defstate
           )
  )

;;;----------------------------------------------------- (EXP) EXPLAIN-ENTRY-STATE

(defstate 
    _exp-entry-state
    (:label "Explain dialog entry")
  (:entry-state _explain-conversation)
  (:explanation "The data is cleaned prior to look for an entry point.")
  (:transitions
   ;;remove useless words explain, define, or else...
   (:patterns ((* "définir" ?y)
               (* "définition" ?y)
               (* "expliquer" ?y)
               (* "expliquez" ?y)
               (* "que" "veut" "dire" ?y)
               ("c" "est" "quoi" ?y "?")
               (?y "c" "est" "quoi" *)
               )
              :keep ?y ; put a list of the binding value into context HAS-DATA
              :sub-dialog _print-concept-documentation-conversation)
   ;; we do not know how the user got here and keep everything
   (:otherwise :sub-dialog _print-concept-documentation-conversation)
   )
  )

;;;================================================================================
;;;
;;;                      GET ADDRESS CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to print the address of a person or of a
;;; place. It fills the task pattern and sends the message to the address specialist.
;;; The simplest version is to send to the ADDRESS agent, wait for the result and
;;; print it.

(defsubdialog 
    _get-address-conversation 
    (:label "Get address conversation")
  (:explanation "Master is trying to obtain an address.")
  (:states _gad-entry-state
           _gad-dont-understand
           _gad-try-again
           _gad-sorry)
  )

;;;================================================================================
;;;                        GET ADDRESS CONVERSATION STATES
;;;================================================================================

;;;----------------------------------------------------- (GAD) ADDRESS-ENTRY-STATE

;;; since we are not sending along any pattern, we will get a list of strings, each
;;; representing an address.

(defstate 
    _get-address-entry-state
    (:entry-state _get-address-conversation)
  (:label "Get address dialog entry")
  (:explanation "Assistant is sending a free-style message to the ADDRESS agent.")
  (:send-message :to :ARS-ADDRESS :self PA_ARISTIDE :action :get-address
                 :args `(((:data . ,(moss::read-fact moss::conversation :input))
                          (:language . :fr))))
  (:transitions
   (:on-failure :target _gad-dont-understand)
   (:otherwise
    :display-answer :success))
  )

;;;----------------------------------------------------- (GAD) GAD-DONT-UNDERSTAND

(defstate 
    _gad-dont-understand
    (:label "Did not understand. Ask master.")
  (:explanation "could not find the address with the given information. Thus, ~
                 asking master for whose address...")
  (:question-no-erase "~%- L'adresse de qui ?")
  (:answer-type :answer)
  (:transitions
   (:always :target _gad-try-again)))

;;;--------------------------------------------------------------- (GAD) GAD-SORRY

(defstate 
    _gad-sorry
    (:label "Address failure.")
  (:explanation "We don't have the requested address.")
  (:text "Désolé, si vous recherchiez une adresse, je ne la trouve pas.~%")
  (:reset)
  (:transitions (:failure))
  )

;;;----------------------------------------------------------- (GAD) GAD-TRY-AGAIN

(defstate 
    _gad-try-again
    (:label "Ask again for address.")
  (:explanation "Assistant is asking ADDRESS staff agent.")
  (:send-message :to :ARS-ADDRESS :self PA_ARISTIDE :action :get-address
                 :args `(((:data . ,(read-fact moss::conversation :input))
                          (:language . :fr))))
  (:transitions
   (:on-failure :target _gad-sorry)
   (:otherwise :display-answer :success))
  )

;;;================================================================================
;;;
;;;                      GET HOME ADDRESS CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to print the home address of a person.
;;; It builds a task pattern and sends the message to the address specialist.
;;; The simplest version is to send to the ADDRESS agent, wait for the result and
;;; print it.

(defsubdialog 
    _get-home-address-conversation 
    (:label "Get address conversation")
  (:explanation "Master is trying to obtain a private address.")
  (:states _ghad-entry-state
           _ghad-dont-understand
           _ghad-process-answer
           _ghad-try-again
           _ghad-try-again-process-answer
           _ghad-sorry)
  )

;;;================================================================================
;;;                 GET HOME ADDRESS CONVERSATION SERVICE FUNCTIONS
;;;================================================================================

;;------------------------------------------------------------ PRINT-HOME-ADDRESS

(defUn print-home-address (addr-list)
  "prints a list of names and addresses"
  ;(format t "~&+++ ARISTIDE/ print-address language: ~S address:~% ~S" 
  ;         *language* addr-list)
  (omas::assistant-display-text 
   PA_ARISTIDE
   (format nil "~{~A~^~2%~}~%" 
     (mapcar #'print-single-home-address addr-list))))

;;;----------------------------------------------------- PRINT-SINGLE-HOME-ADDRESS
;;; Each entry is like 
;;;   (("name" "Barthès")("first name" "Jean-Paul")("home address" "..."))
;;; Home address possibly empty

(defUn print-single-home-address (addr)
  "returns a single charater string"
  (format nil "~{~A~^ ~}: ~{~A~^ ~}~%~{~A~^ ~}"
    ;; get person name
    (or (cdr (assoc "nom" (cdr addr) :test #'string-equal)) 
        (list "?"))
    ;; name
    (or (cdr (assoc "prénom" (cdr addr) :test #'string-equal)) 
        (list "?"))
    ;; address
    (let ((data (cdr (assoc "adresse" (cdr addr) :test #'string-equal))) 
          result)
      (if (setq result (assoc "adresse privée" data :test #'string-equal))
          (cdr result)
        (list "adresse du domicile inconnue...")))))

#|
? (print-single-home-address '("personne"
                                 ("nom" "Barthès")("prénom" "Camille")
                                      ("adresse privée")))
"Barthès: Dominique
adresse du domicile inconnue...
"
? (print-single-home-address '("personne" ("nom" "Barthès") ("prénom" "Jean-Paul")
            ("adresse"
             ("adresse privée"
              "14, Allée de la Tilloye, 60200 Compiègne, (France)"))))
|#
;;;================================================================================
;;;                    GET HOME ADDRESS CONVERSATION STATES
;;;================================================================================		   

;;;------------------------------------------- (GHAD) GET-HOME-ADDRESS-ENTRY-STATE

(defstate 
    _get-home-address-entry-state
    (:entry-state _get-home-address-conversation)
  (:label "Get home address dialog entry")
  (:explanation "Assistant is sending a free-style message to the ADDRESS agent.")
  (:send-message :to :ARS-ADDRESS :self PA_ARISTIDE :action :get-address
                 :args `(((:data . ,(read-fact moss::conversation :input))
                          (:language . :fr)
                          (:pattern . ("personne" 
                                       ("nom")
                                       ("prénom")
                                       ("adresse" ("adresse privée")))))))
  (:transitions
   (:on-failure :target _ghad-dont-understand)
   (:otherwise
    :print-answer #'print-home-address :success))
  )

;;;-------------------------------------------------------- (GHAD) DONT-UNDERSTAND

(defstate 
    _ghad-dont-understand
    (:label "Did not understand. Ask master.")
  (:explanation "could not find the address with the given information. Thus, ~
                 asking master for whose address...")
  (:question "~%- L'adresse personnelle de qui ?")
  (:answer-type :answer)
  (:transitions
   (:always :target _ghad-try-again)))

;;;------------------------------------------------------------------ (GHAD) SORRY

(defstate 
    _ghad-sorry
    (:label "Address failure.")
  (:explanation "We don't have the requested address.")
  (:text 
   "Désolé, si vous recherchiez une adresse personnelle, je ne la trouve pas.~%")
  (:reset)
  (:transitions (:always :failure))
  )

;;;-------------------------------------------------------------- (GHAD) TRY-AGAIN

(defstate 
    _ghad-try-again
    (:label "Ask again for address.")
  (:explanation "Assistant is asking ADDRESS staff agent.")
  (:send-message :to :ARS-ADDRESS :self PA_ARISTIDE :action :get-address
                 :args `(((:data . ,(read-fact moss::conversation :input))
                          (:language . :fr)
                          (:pattern . ("personne" 
                                       ("nom")
                                       ("prénom")
                                       ("adresse" ("adresse privée")))))))
  (:transitions
   (:on-failure :target _ghad-sorry)
   (:otherwise :print-answer #'print-home-address :success)
   )
  )

;;;===============================================================================
;;;
;;;                       GET LAST NEWS CONVERSATION
;;;
;;;===============================================================================

(defsimple-subdialog "Get-last-news" "_gln"
  :explanation "On désire les dernières nouvelles."
  :from :ARISTIDE :to :NEWS :action :get-last-news
  :language :fr
  :sorry "- Je n'ai pas de nouvelles en stock")


;;;================================================================================
;;;
;;;                      GET NEWS BY SECTION CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to print news about a specific topic. It fills 
;;; the task pattern and sends the message to the news specialist.

(defsubdialog 
    _get-news-by-section-conversation 
    (:label "Get news by section conversation")
  (:explanation "Master is trying to obtain news of a specific section.")
  (:states _gnbsd-entry-state
           _gnbsd-dont-understand
           _gnbsd-try-again
           _gnbsd-sorry)
  )

;;;------------------------------------------------------- (GNBSD) GNBS-ENTRY-STATE

(defstate 
    _get-news-by-section-entry-state
    (:entry-state _get-news-by-section-conversation)
  (:label "Get news by section dialog entry")
  (:explanation "Assistant is sending a free-style message to the NEWS agent.")
  (:send-message :to :NEWS :self PA_ARISTIDE :action :get-news-by-section
                 :args `(((:data . ,(moss::read-fact moss::conversation :input))
                          (:language . :fr))))
  (:transitions
   (:on-failure :target _gnbsd-dont-understand)
   (:otherwise :display-answer :success))
  )

;;;--------------------------------------------------- (GNBSD) GNBS-DONT-UNDERSTAND

(defstate 
    _gnbsd-dont-understand
    (:label "Did not understand. Ask master.")
  (:explanation "could not find news in the requested section. Thus, ~
                 asking master for what section...")
  (:question-no-erase "~%- Vous voulez des nouvelles sur quelle rubrique ?")
  (:answer-type :answer)
  (:transitions
   (:always  :target _gnbsd-try-again)))

;;;------------------------------------------------------------------ (GNBSD) SORRY

(defstate 
    _gnbsd-sorry
    (:label "News obtention failure.")
  (:explanation "We don't have the requested news.")
  (:text "Desolé, je n'ai rien trouvé.~%")  
  (:reset)
  (:transitions (:failure))
  )

;;;-------------------------------------------------------------- (GNBSD) TRY-AGAIN

(defstate 
    _gnbsd-try-again
    (:label "Ask again for news.")
  (:explanation "Assistant is asking :NEWS.")
  (:send-message :to :NEWS :self PA_ARISTIDE :action :get-news-by-section
                 :args `(((:data . ,(moss::read-fact moss::conversation :input))
                          (:language . :fr)
                          )))
  (:transitions
   (:on-failure :target _gnbsd-sorry)
   (:otherwise :display-answer :success))
  )

;;;===============================================================================
;;;
;;;                       GET NEWS SECTIONS CONVERSATION
;;;
;;;===============================================================================

(defsimple-subdialog "Get-news-sections" "_gns"
  :explanation "On désire la liste des catégories de nouvelles."
  :from :ARISTIDE :to :NEWS :action :get-news-sections
  :language :fr
  :sorry "- Je n'ai pas trouvé la liste des catégories demandée.")

;;;================================================================================
;;;
;;;                          GET PHONE NUMBER CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to print the telephone number of a person or of a
;;; place. It fills the task pattern and sends the message to the address specialist.
;;; The simplest version is to send to the ADDRESS agent, wait for the result and
;;; print it.

;; create a T-CONTEXT from the task description. The T-CONTEXT contains the dynamic
;; information corresponding to the particular instance of this conversation.
;; Could we put this info into the Q-CONTEXT with additional properties?

;; fill it with the arguments

;; put the rest of the data into the reouired data argument

;; ship it to the specialist

;; wait to see what happens

(defsubdialog 
    _get-phone-conversation
    (:label "Get phone dialog")
  (:explanation "Master is trying to obtain a telephone number.")
  (:states _gpd-entry-state
           _gpd-dont-understand
           _gpd-process-answer
           _gpd-process-second-answer
           _gpd-try-again
           _gpd-sorry)
  )

;;;================================================================================
;;;                    GET TELEPHONE CONVERSATION STATES
;;;================================================================================

;;;------------------------------------------------------------- (GPD) ENTRY-STATE

(defstate 
    _gpd-entry-state
    (:label "Get phone dialog")
  (:entry-state _get-phone-conversation)
  (:explanation "Assistant is sending a free-style message to the ADDRESS agent.")
  (:send-message :to :ARS-ADDRESS :self PA_ARISTIDE :action :get-phone-number
                 :args `(((:data . ,(read-fact moss::conversation :input))
                          (:language . :fr))))
  (:transitions
   (:on-failure :target _gpd-dont-understand)
   (:otherwise :display-answer :success))
  )

;;;--------------------------------------------------------- (GPD) DONT-UNDERSTAND

(defstate 
    _ghad-dont-understand
    (:label "Did not understand. Ask master.")
  (:explanation "could not find the address with the given information. Thus, ~
                 asking master for whose address...")
  (:question "~%- L'adresse personnelle de qui ?")
  (:answer-type :answer)
  (:transitions
   (:always :save :target _ghad-try-again)))

;;;------------------------------------------------------------------- (GPD) SORRY

(defstate 
    _gpd-sorry
    (:label "Phone number failure.")
  (:explanation "We don' have the requested number.")
  (:reset)
  (:text "- Désolé, je ne trouve pas le numéro demandé.")
  (:transitions (:failure))
  )

;;;--------------------------------------------------------------- (GPD) TRY-AGAIN

(defstate 
    _gpd-try-again
    (:label "Ask again for phone number.")
  (:explanation "Assistant is asking ADDRESS.")
  (:send-message :to :ARS-ADDRESS :self PA_ARISTIDE :action :get-phone-number
                 :args `(((:data . ,(read-fact moss::conversation :input))
                          (:language . :fr))))
  (:transitions
   (:on-failure :target _gpd-sorry)
   (:otherwise :display-answer :success))
  )

;;;================================================================================
;;;
;;;                  PRINT CONCEPT DOCUMENTATION CONVERSATION
;;;
;;;================================================================================

;;; this conversation is meant to print the definition of a concept or properties
;;; to be extracted from a list of words

(defsubdialog
    _PRINT-CONCEPT-DOCUMENTATION-CONVERSATION
    (:label "Print concept documentation")
  (:explanation 
   "This dialog is meant to print the documentation attached to concepts or ~
    properties referenced by a list of words.")
  (:states _pcd-entry-state
           _pcd-sorry)
  )

;;;-------------------------------------------------------------- (PCD) ENTRY-STATE

(defstate 
    _pcd-entry-state
    (:label "start of the PRINT CONCEPT DOCUMENTATION dialog")
  (:entry-state _print-concept-documentation-conversation)
  (:explanation 
   "Initial state may have data left in the input area. The data either represent ~
    an entry point or something else that should be used to select objects.")
  (:answer-analysis)
  )

(defownmethod
    =answer-analysis _pcd-entry-state (conversation input)
  "Using list of words to access objects. If one, OK, if more, must ask user to ~
   select one, if none, failure."
  ;(declare (ignore more-args))
  (let* ((*language* :fr)
         entry-point-list object-list)
    (catch :return
           (moss::vformat "_pcd-entry-state /input: ~%  ~S" input)
           ;; first get entry-points
           (setq entry-point-list (moss::find-best-entries input))
           
           ;; if we cannot find anything locally try for a what-is conversation
           (unless entry-point-list
             (throw :return 
               `(:sub-dialog _wh?t-is-conversation :failure ,_pcd-sorry)))
           
           ;; get objects corresponding to entry points
           ;(moss::vformat "_pcd-entry-state /entry-point-list: ~%  ~S" entry-point-list)
           (setq object-list (delete-duplicates (mapcan #'access entry-point-list)))
           (moss::vformat "_pcd-entry-state /object-list: ~%  ~S" object-list)
           
           ;; first filter concepts and properties
           (setq object-list
                 (mapcan #'(lambda (xx)
                             (or 
                              (moss::%is-concept? xx)
                              (moss::%is-attribute? xx)
                              (moss::%is-relation? xx)
                              )
                             (list xx))
                   object-list))
           ;(print object-list)
           (if object-list
               ;; print documentation
               (dolist (obj-id object-list)
                 (cond
                  ((moss::%is-concept? obj-id)
                   (send obj-id '=get-documentation :lead "CONCEPT: " 
                         :final-new-line t)
                   (send conversation '=display-text *answer*))
                  ((moss::%is-attribute? obj-id)
                   (send obj-id '=get-documentation :lead "ATTRIBUT: " 
                         :final-new-line t)
                   (send conversation '=display-text *answer*))
                  ((moss::%is-relation? obj-id)
                   (send obj-id '=get-documentation :lead "RELATION: " 
                         :final-new-line t)
                   (send conversation '=display-text *answer*))
                  )
                 )
             ;; otherwise print failure
             (throw :return 
               `(:transitions ,_pcd-sorry))
             ;(send conversation '=display-text 
             ;      "~%Désolé, je n'ai pas de concept ni de propriété correspondant ~
             ;       à votre demande.~%")
             )
           ;; success
           (list :success))))

;;;------------------------------------------------------------------- (PCD) SORRY

(defstate 
    _pcd-sorry
    (:label "Reference search failure.")
  (:explanation "We did not find any explanation for the reference.")
  (:reset)
  (:text "~%- Désolé, je n'ai pas de concept ni de propriété correspondant ~
          à votre demande.~%." :no-erase)
  (:transitions 
   (:failure))
  )

;;;================================================================================
;;;
;;;                       PRINT HELP CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to help the master by giving information:
;;;   - in general (what tasks are available
;;;      "what can you do for me?" "help." "what can I do?"
;;;   - on a particular topic
;;;      "how do I send a mail?" "help me with the mail?"
;;; specific help corresponding to special help tasks are caught by the task 
;;; task selection mechanism?

(defsubdialog 
    _print-help-conversation 
    (:label "Help conversation")
  (:explanation 
   "Help was asked.")
  (:states _ph-print-global-help)
  )

;;;-------------------------------------------------------- (PH) PRINT-GLOBAL-HELP

(defstate
    _ph-print-global-help
    (:label "Print general help")
  (:entry-state _print-help-conversation )
  (:explanation "No specific subject was included.")
  (:execute
   (let
       ((obj-id (car (send '>-global-help '=get 'is-title-of)))
        (*language* :fr)
        (task-list (access '(task)))
        (conversation (omas::conversation aristide::PA_ARISTIDE))
        )
     (when obj-id (send obj-id '=get-documentation)
       (send conversation '=display-text *answer*))
     (send conversation '=display-text
           "~2%Je peux faire les choses suivantes :")
     (dolist (task task-list)
       (send task '=get-documentation :lead "   - ")   
       (send conversation '=display-text *answer*)
       )
     (send conversation '=display-text "~2%")
     ))
  (:transitions
   (:always :success))
  )


;;;================================================================================
;;;
;;;                         PROCESS ASK CONVERSATION
;;;
;;;================================================================================

;;; this conversation is meant to post a message for inspection and send a possible 
;;; answer, or abort the corresponding task

;;; The problem is that the PA has received a request corresponding to the message
;;; being processed and the static part of the skill has posted the message and
;;; is waiting for the answer.
;;; One problem is that there are as many threads as posted messages waiting for an 
;;; answer.

;;; If we went to answer the request message, then the situation is quite complex:
;;; we have several processes:
;;;   - the thread executing the ask skill
;;;   - the converse process
;;;   - the assistant window thread handling the callbacks
;;;   - the timeout thread on the ask thread
;;; One can solve all the problems but the result is not simple.
;;;
;;; A better approach is to consider that all messages ASk and TELL are considered
;;; as INFORM messages, and do not create any specific process (i.e. waiting for a
;;; master's answer), à la e-mail. This allows to send answers directly as INFORM
;;; messages, bypassing the complex mechanism of getting the control back to the
;;; ASK skill.
;;; Thus, all request messages triggering the ASK skill must send a response like
;;; "message well received, transmitted to the master." 

;;; In the simplified approach the role of the PROCESS ASK conversation is to answer
;;; the message when the PROCESS (TRAITER) button is clicked.
;;; The sequence of events is:
;;;   - the ":process-ask" tag is inserted into the agent TO-DO slot
;;;   - the converse process wakes up and triggers the PROCESS ASK dialog
;;;   - a message is posted to the master
;;; If the master wants to answer then
;;;   - the receiver is first determined
;;;   - the *pass-every-char* flag (that should be a specific slot of the interface
;;;     window) is set
;;;   - the master inputs text
;;;   - the DONE (Terminé) button is clicked resetting *pass-every-char* 
;;;   - the text is transferred into the :input area of FACTS
;;;   - the message is sent to the receiver directly as an inform message

(defsubdialog 
    _process-ask-conversation
    (:label "Process ask dialog")
  (:explanation "Display the content of a question message (ask).")
  (:states  _pa-entry-state  ; required by defstate
           _pa-brush ; clean up lightly
           _pa-destroy ; destroy message return :abort to ask-static
           _pa-discard? ; ask whather we dump the question or not
           _pa-get-answer ; get the answer from the master
           )
  )

;;;=========================== service functions ==================================

;;;------------------------------------------------------ (PA) DISPLAY-ASK-MESSAGES

(defUn display-ask-message (message conversation)
  "takes a message and displays its content into the assistant pane"
  (let (sender content)
    ;; get sender identity
    (setq sender (omas::from! message))
    ;; get content of the message
    (setq content (car (omas::args message)))
    ;; ask to print it
    (send conversation '=DISPLAY-TEXT 
          (format nil "Expéditeur : ~A ~%Urgence : ~A ~%Objet : ~A ~2% ~A"
            sender
            (or (cadr (member :priority content)) "Normal")
            (or (cadr (member :object content)) "")
            (or (getf content :text) "<empty message>")))))

;;;================================================================================

;;;-------------------------------------------------------------- (PA) ENTRY-STATE
;;; We enter the dialog when the PROCESS (Traiter) button has been clicked. 
;;; FACTS/INPUT contains ":proces-ask" and message is in FACTS/TODO-MESSAGE

(defstate 
    _pa-entry-state
    (:label "process ask dialog")
  (:entry-state _process-ask-conversation)
  (:explanation "Shows the message and ask if an answer is needed.")
  (:execute-preconditions
   ;; the :input area contains a pair (<message> . <process-id>)
   (display-ask-message (read-fact moss::conversation :todo-message)
                        moss::conversation))
  (:question-no-erase "~2%Voulez-vous répondre à ce message?")
  (:transitions 
   (:yes :target _pa-get-answer) 
   (:otherwise :target _pa-discard?))
  )

;;;-------------------------------------------------------------------- (PA) BRUSH
;;; reset conversation slot
(defstate 
    _pa-brush
    (:label "bookkeeping")
  (:explanation "do some light bookkeeping.")
  (:execute (replace-fact moss::conversation :todo-message nil))
  (:transitions
   (:always :success))
  )

;;;------------------------------------------------------------------ (PA) DESTROY


(defstate 
    _pa-destroy
    (:label "Destroy ASK message.")
  (:explanation "Master does not want to answer the message nor keep it. We ~
                 remove it from the list and return to the task with an :abort~
                 mark, so that it will exit without answering.")
  (:transitions
   (:always
    :exec 
    (let ((agent (car (has-moss-agent moss::conversation))))
      ;; discard the message from the list, assuming it is still selected
      (omas::ASSISTANT-DISCARD-SELECTED-TO-DO-TASK agent)
      ;; clean up conversation
      (replace-fact moss::conversation :todo-message nil))
    :success))
  )

;;;----------------------------------------------------------------- (PA) DISCARD?

(defstate 
    _pa-discard?
    (:label "ask whether we throw the message away.")
  (:explanation "Master did not answer message, ask master if we keep it to it ~
                 later.")
  (:question-no-erase "~2%Voulez-vous garder ce message pour le traiter plus tard ?")
  (:transitions
   (:no :target _pa-destroy)
   (:otherwise :target _pa-brush))
  )

;;;--------------------------------------------------------------- (PA) GET-ANSWER

(defstate 
    _pa-get-answer
    (:label "obtain master's answer to an ask message.")
  (:explanation "Record text typed into the master pane until the DONE button is ~
                 pushed. We must disable . and ? terminations.")
  ;; set global so that we can have . and ? in the text
  (:execute-preconditions 
   (let ((win (car (has-moss-input-window moss::conversation))))
     (when win
       (setf (omas::pass-every-char win) t)))
   )
  (:question-no-erase "~% ...Tapez votre réponse puis cliquez sur le bouton Terminer.") 
  (:answer-analysis)
  )

(defownmethod =answer-analysis _pa-get-answer (conversation input)
  "We process the answer and send it back unless we get an abort.
Arguments:
   conversation: conversation object is
   input: copy of the :input area of FACTS"
  (let (agent message content object answer)
    (setq agent (car (has-moss-agent conversation)))
    ;(print  `(=====> input ,input))
    ;; if input is different from :abort we build a fuller answer
    (unless (eql input :abort)
      (setq message (read-fact conversation :todo-message)
          content (car (omas::args message))
          object (getf content :object))
      ;; format an answer content adding object to text
      (setq answer (list :object (format nil "RE: ~A" (or object " "))	
                         :text input))
      ;; make an answer and send it back as an INFORM message
      (send-message 
       (make-instance 'omas::message
         :from :ARISTIDE
         :to (omas::from! message)
         :date (get-universal-time)
         :type :inform
         :action :tell
         :args (list answer)
         )
       )
      )
    ;; clean FACTS
    (replace-fact conversation :todo-message nil)
    ;; remove message from the tasks to do pane
    (omas::ASSISTANT-DISCARD-SELECTED-TO-DO-TASK agent)
    )
  ;; we then get out of dialog
  '(:success))

;;;================================================================================
;;;
;;;                         PROCESS TELL CONVERSATION
;;;
;;;================================================================================

;;; this conversation is meant to post a message for inspection and quit

(defsubdialog 
    _process-tell-conversation
    (:label "Process tell dialog")
  (:explanation "Display the content of an info message (tell).")
  (:states _pt-entry-state  ; required by defstate
           _pt-erase? ; go clean the message
           _pt-clean ; remove item from list
           _pt-brush ; clean up lightly
           )
  )

;;;=========================== service functions ==================================

;;;----------------------------------------------------- (PT) DISPLAY-TELL-MESSAGES

(defUn display-tell-message (message conversation)
  "takes a message and displays its content into the assistant pane"
  (let (sender content)
    ;; get sender identity
    (setq sender (omas::from! message))
    ;; get content of the message
    (setq content (car (omas::args message)))
    ;; ask to print it
    (send conversation '=DISPLAY-TEXT 
          (format nil "Expéditeur : ~A ~%Urgence : ~A ~%Objet : ~A ~2% ~A"
            sender
            (or (cadr (member :priority content)) "Normal")
            (or (cadr (member :object content)) "")
            (or (getf content :text) "<empty message>")))))

;;;================================================================================
;;;                        PROCESS TELL CONVERSATION STATES
;;;================================================================================

;;;-------------------------------------------------------------- (PT) ENTRY-STATE

(defstate 
    _pt-entry-state
    (:label "Process tell dialog")
  (:entry-state _process-tell-conversation)
  (:explanation "Affiche un message et demande si on le garde.")
  (:execute (display-tell-message 
             (car (read-fact moss::conversation :message)) MOSS::Conversation))
  (:question-no-erase "~2%OK? J'efface le message?")
  (:transitions 
   (:yes :target _pt-clean) 
   (:otherwise :target _pt-brush))
  )

;;;-------------------------------------------------------------------- (PT) BRUSH

(defstate 
    _pt-brush
    (:label "bookkeeping")
  (:explanation "remove item from facts.")
  (:execute (replace-fact moss::conversation :message nil))
  (:transitions
   (:always :success))
  )

;;;-------------------------------------------------------------------- (PT) CLEAN

(defstate 
    _pt-clean
    (:label "Cleaning state")
  (:explanation "Ask whether we keep the message in the to-do pane.")
  (:execute (omas::ASSISTANT-DISCARD-SELECTED-TO-DO-TASK 
             (car (has-moss-agent moss::conversation)))
            (replace-fact moss::conversation :message nil))
  (:transitions
   (:always :success))
  )

;;;================================================================================
;;;
;;;                      PUBLISH NEWS CONVERSATION
;;;
;;;================================================================================

(defsubdialog 
    _publish-new-conversation
    (:label "Publish new conversation")
  (:explanation 
   "Master is trying to publish a new.")
  (:states _publish-new-entry-state
           _pnd-get-category
           _pnd-get-content
           _pnd-get-sections
           _pnd-publish
           _pnd-sorry)
  )

;;;------------------------------------------------------------- (PND) ENTRY-STATE

(defstate 
    _publish-new-entry-state
    (:entry-state _publish-new-conversation)
  (:label "Publish new dialog entry")
  (:explanation "Ask for the title of the new.")  
  (:question-no-erase "~%- Quelle est le sujet de la nouvelle ?")
  (:answer-type :answer)
  (:execute 
   (omas::remember PA_ARISTIDE (moss::read-fact moss::conversation :input)
                   :news-title))
  (:transitions
   (:always  :target _pnd-get-category)))

;;;------------------------------------------------------------ (PND) GET-CATEGORY

(defstate 
    _pnd-get-category
    (:label "Ask master the content.")
  (:explanation "Obtain title, ask for category")
  ;(format t "~S~%" (moss::read-fact moss::conversation :input))
  ;(format t "~% TITULO A GUARDAR: ~S" (moss::read-fact moss::conversation :input))
  ;; Store title in agent´s short term memory  OMARAGP PORQUE ESTO NO FUNCIONA  
  (:question-no-erase "~%- Dans quelle rubrique voulez-vous enregistrer la nouvelle ?")
  (:execute 
   (omas::remember PA_ARISTIDE (moss::read-fact moss::conversation :input)
                   :news-section))
  (:transitions
   (:patterns ((?* ?x) "rubriques" (?* ?y))
              :target _pnd-get-sections)
   (:otherwise  :save :target _pnd-get-content)))

;;;------------------------------------------------------------- (PND) GET-CONTENT

(defstate 
    _pnd-get-content
    (:label "Ask master the content.")
  (:explanation "Obtain category, ask for content")
  
  (:execute-preconditions 
   (setf (omas::pass-every-char (car (has-moss-input-window moss::conversation))) t)
   ;; include :verbatim flag into the FACT base (?)
   (replace-fact moss::conversation :verbatim t)
   )
  (:question-no-erase "~%- Ecrivez le contenu de la nouvelle et puis cliquez sur terminer")
  (:answer-type :answer)
  (:execute 
   ;; Store category in agent´s short term memory						   
   (remember PA_ARISTIDE (moss::read-fact moss::conversation :input)
             :news-content))
  (:transitions
   (:always  :target _pnd-publish)))

;;;------------------------------------------------------------ (PND) GET-SECTIONS

(defstate 
    _pnd-get-sections
    (:label "Ask NEWS about sections.")
  (:explanation "Master does not know sections. Asking NEWS for the list.")
  (:send-message :to :NEWS :self PA_ARISTIDE :action :get-all-sections
                 :args '((:language . :fr)))
  (:execute 
   (omas::assistant-display-text
    PA_ARISTIDE
    (format nil "~S" (moss::read-fact moss::conversation :answer))))
  (:transitions
   (:on-failure :target _pnd-sorry)
   (:otherwise
    ;:print-answer #'print-sections :target _pnd-get-category)
    :target _pnd-get-category)))

;;;----------------------------------------------------------------- (PND) PUBLISH
;;; to publish news we send a structure that can be used in a defindividual clause

(defstate 
    _pnd-publish
    (:label "Ask master the content.")
  (:explanation "Ask NEWS to create the news")
  (:send-message :to :NEWS :self PA_ARISTIDE :action :publish-new
                 ;; Recall gets data stored un ALBERT´s short term memory
                 :args `((:data
                          "news"
                          ("sujet" ,(recall PA_ARISTIDE :news-title))
                          ("section" ,(recall PA_ARISTIDE :news-section))
                          ("contenu" ,(recall PA_ARISTIDE :news-content))
                          ("autheur"  "Aristide")
                          (:language . :fr))))
  ;(:execute (print (car (READ-FACT MOSS::CONVERSATION :INPUT))))
  (:transitions
   (:on-failure :target _pnd-sorry)
   (:otherwise :success))
  )

#|
(defownmethod =resume _pnd-publish (conversation &rest more-args)
  "we got an answer from the NEWS message or else the master got impatient.
Arguments:
   agent: assistant."
  (declare (ignore more-args))
  (let* ((agent (car (has-moss-agent conversation)))
         (answer (omas::answer agent))
         (*language* :fr)
         )
    ;(format *debug-io* "~&+++ GLN =resume; HAS-TO-DO: ~S" (has-to-do conversation))    
    
    (moss::vformat "_gln-entry-state /Answer to the send-request:~& ~S" answer)
    ;; on failure do something
    (cond     
     ((equal "failure" answer)
      `(:transition ,_pnd-sorry))
     (t 
      `(:success)))))
|#

;;;---------------------------------------------------------------- (PND) PND-SORRY

(defstate 
    _pnd-sorry
    (:label "Publish failure.")
  (:explanation "It was an error while publishing the new.")
  (:text "Desolé, l'agent NEWS n'a pas pu publier votre texte.~%")
  (:reset)
  ;(:immediate-transitions (:failure))
  (:transitions (:failure))
  ;(:no-resume)
  )

;;;================================================================================
;;;
;;;                       SET FONT SIZE CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to modify the size of the dialog font, usually to
;;; increase it or to return to the default value.

(defsubdialog 
    _set-font-size-conversation
    (:short-name sfs)
  (:label "Font size conversation")
  (:explanation 
   "The data is searched for the presence of words like increase, ~
    bigger, smaller, defaults, reset,...")
  (:states _sfs-entry-state  ; required by defstate
           _sfs-dont-understand
           _sfs-try-again
           _sfs-sorry)
  ;(:action _sfs-action)
  )

;;;--------------------------------------------------- (SFS) FONT-SIZE-ENTRY-STATE

(defstate 
    _sfs-entry-state
    ;(:conversation _set-font-size-conversation)
    (:label "Font size dialog entry")
  (:entry-state _set-font-size-conversation)
  (:explanation "The data is searched for the presence of words like increase, ~
                 bigger, defaults, reset,...")
  (:transitions
   (:patterns ((* "augmenter" *)
               (* "plus" "grand" *)
               (* "trop" "petit" *)
               (* "plus" "gros" *))
              :exec (omas::assistant-use-larger-font PA_ARISTIDE) :success)
   (:patterns ((* "diminuer" *)
               (* "plus" "petit" *)
               (* "moins" "grand" *)
               (* "moins" "gros" *))
              :exec (omas::assistant-reset-font PA_ARISTIDE) :success)
   (:patterns ((* "utiliser" * "défaut" *)
               (* "reset" *)
               (* "défaut" *))
              :exec (omas::assistant-reset-font PA_ARISTIDE) :success)
   (:otherwise :target _sfs-dont-understand)
   )
  )

;;;--------------------------------------------------------- (SFS) DONT-UNDERSTAND

(defstate 
    _sfs-dont-understand
    (:label "Did not understand. Ask master.")
  (:explanation "Assistant is asking master for font-size clarification.")
  (:question "~%- Vous voulez que j'écrive plus gros, plus petit, ou que ~
              j'utilise la taille par défaut ?")
  (:transitions
   (:always :target _sfs-try-again)))

;;;------------------------------------------------------------------- (SFS) SORRY

(defstate 
    _sfs-sorry
    (:label "Font dialog failure.")
  (:explanation "We cannot make things out from the data.")
  (:reset)
  (:text "~%- Désolé, je ne vois pas ce que vous voulez faire avec les lettres.~%")
  (:transitions
   (:always :failure))
  )

;;;--------------------------------------------------------------- (SFS) TRY-AGAIN

(defstate 
    _sfs-try-again
    (:label "Font size try again")
  (:explanation "we asked master and try to pick up the answer.")
  (:transitions
   (:patterns ((* "augmenter" *)
               (* "plus" "grand" *)
               (* "plus" "gros" *))
              :exec (omas::assistant-use-larger-font PA_ARISTIDE) :success)
   (:patterns ((* "diminuer" *)
               (* "plus" "petit" *)
               (* "moins" "grand" *)
               (* "moins" "gros" *))
              :exec (omas::assistant-reset-font PA_ARISTIDE) :success)
   (:patterns ((* "utiliser" * "défaut" *)
               (* "reset" *)
               (* "défaut" *)
               (* "defaut" *))
              :exec (omas::assistant-reset-font PA_ARISTIDE) :success)
   (:otherwise :target _sfs-sorry)
   )
  )
;;;================================================================================
;;;
;;;                           TELL CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to tell something to a remote PA.

;;; We must first locate our receiver (presumably the PA of somebody in the coterie
;;; We then want to prepare topic and text, the send the message

(defsubdialog 
    _tell-conversation 
    (:label "Tell conversation")
  (:explanation "We want our PA to tell something to somebody.")
  (:states _tell-entry-state  ; required by defstate            
           _tell-choose-destination 
           _tell-get-data
           _tell-get-destination
           _tell-send-message
           _tell-sorry
           )
  )

;;;================================================================================
;;;                       TELL CONVERSATION SERVICE FUNCTIONS
;;;================================================================================

;;;------------------------------------------------------------ EXTRACT-DESTINATION
;;; See ASK dialog

;;;================================================================================
;;;                            TELL CONVERSATION STATES
;;;================================================================================

;;;-------------------------------------------------------- (TELL) TELL-ENTRY-STATE

(defstate 
    _tell-entry-state
    (:label "Tell dialog entry")
  (:entry-state _tell-conversation)
  (:explanation "We try to get destination from the input data.")
  (:answer-analysis)
  )


(defownmethod
    =answer-analysis _tell-entry-state (conversation input)
  "Using list of words to locate receiver of message to send. Must ask user ~
   to confirm findings.
Arguments:
   conversation: conversation object
   input copy of the FACTS/input area"
  (let* (bindings destination)
    (catch 
     :return
     (cond
      ;; when input is empty then failure
      ((null input)
       (throw :return (list :failure)))
      ;; try to extract destination info from the input text
      ((setq bindings
             (MOSS::PROCESS-PATTERNS 
              input
              '(((?* ?X) "dire" "à" (?* ?Y)) 
                ((?* ?X) "informer" (?* ?Y))
                ((?* ?X) "faire" "savoir" "à" (?* ?Y))
                ((?* ?X) "message" "pour" (?* ?Y)))))
       (setq input (cdr (assoc '?Y bindings)))
       )
      ;; otherwise do nothing
      )
     
     ;; try to get destination
     (setq destination (extract-destination input))
     
     (cond
      ;; if null, failure
      ((null destination)
       (throw :return (list :transition _tell-sorry)))
      ;; list, must make a choice
      ((listp destination)
       ;; save the result into the FACT area and go ask for choice
       (replace-fact conversation :to-choose destination)
       (throw :return (list :transition _tell-choose-destination)))
      ;; otherwise we got destination, save it
      (t
       (replace-fact conversation :to destination)
       ;; and go get data
       (throw :return (list :transition _tell-get-data))))
     )))

;;;------------------------------------------------------ (TELL) CHOOSE-DESTINATION

(defstate 
    _tell-choose-destination
    (:label "Tell choose destination")
  (:explanation "ambiguous destination (list) select one.")
  (:execute
   (let* ((conversation moss::conversation)
          (choices (read-fact conversation :to-choose))
          )
     ;; remove the temporary data
     (replace-fact conversation :to-choose nil)
     (replace-fact conversation :to (car choices))))
  (:transitions (:always :target _tell-get-data)))

;;;--------------------------------------------------------------- (TELL) GET-DATA

(defstate 
    _tell-get-data
    (:label "set up parameters for continuous input.")
  (:explanation "Let the master type text into the pane until the DONE button is ~
                 pushed. We must disable . and ? terminations.")
  ;; set global flag in the interface window so that we can have . and ? in the text
  (:execute-preconditions 
   (setf (omas::pass-every-char (car (has-moss-input-window moss::conversation))) t)
   ;; include :verbatim flag into the FACT base (?)
   (replace-fact moss::conversation :verbatim t)
   )
  (:question-no-erase 
   "~% ...Tapez ou dictez votre texte puis cliquez sur le bouton Terminer.")
  (:transitions
   (:always :target _tell-send-message))
  )

;;;--------------------------------------------------------- (TELL) GET-DESTINATION

(defstate 
    _tell-get-destination
    (:label "obtain master's destination for message.")
  (:explanation "Get the id of the personal assistant of the receiver.")
  ;; set global so that we can have . and ? in the text
  (:question-no-erase 
   "~% A qui voulez-vous envoyer le message ?") 
  (:answer-analysis)
  )

(defownmethod =answer-analysis _tell-get-destination (conversation input)
  "We process the answer and record it unless we get an abort.
Arguments:
   agent: assistant."
  (let (destination)
    ;; analyze input
    (setq destination (extract-destination input))
    
    (cond
     ((null destination)
      (list :transition _tell-sorry))
     ((listp destination)
      (replace-fact conversation :to-choose destination)
      (list :transition _tell-choose-destination))
     (t
      (replace-fact conversation :to destination)
      ;; we then get out of dialog
      (list :transition _tell-get-data)))))

;;;------------------------------------------------------------ (TELL) SEND-MESSAGE

(defstate 
    _tell-send-message
    (:label "Tell send final message")
  (:explanation "build and send the message.")
  (:send-message-no-wait 
   :to (pop-fact moss::conversation :to)
   :type :inform
   :action :tell
   :args (list (list :text (pop-fact moss::conversation :raw-input))))
  (:text "~&...Message envoyé")
  (:transitions (:always :success)) ; return
  )

;;;------------------------------------------------------------------- (TELL) SORRY

(defstate 
    _tell-sorry
    (:label "Tell failure.")
  (:explanation "We don't have the needed parameters")
  (:text 
   "Désolé, je ne vois pas à qui vous voulez envoyer le message.~%")
  (:reset)
  (:transitions (:failure))
  )

;;;================================================================================
;;;
;;;                              TRACE CONVERSATION
;;;   
;;;================================================================================

;;; this conversation is intended to modify the size of the dialog font, usually to
;;; increase it or to return to the default value.

(defsubdialog 
    _trace-conversation
    (:label "Trace conversation")
  (:explanation "Command to toggle the dialog trace.")
  (:states _tr-entry-state)
  )

;;;-------------------------------------------------------- (TR) TRACE-ENTRY-STATE

(defstate 
    _tr-entry-state
    (:label "Trace dialog entry")
  (:entry-state _trace-conversation)
  (:explanation "We search the input for the trace command.")
  
  (:transitions
   ;; system commands
   (:patterns (((?* ?x) "afficher" (?* ?y) "transitions" (?* ?z))
               ((?* ?x) "tracer" (?* ?y) "transitions" (?* ?z))
               ((?* ?x) "imprimer" (?* ?y) "transitions" (?* ?z)))
              :exec (setq moss::*verbose* t) :success)
   (:patterns (((?* ?x) "pas" "de" "trace" (?* ?y))
               ((?* ?x) "supprimer" "la" "trace" (?* ?y))
               ((?* ?x) "arrêter" "la" "trace" (?* ?y))
               ((?* ?x) "stop" "trace" (?* ?y)))
              :exec (setq moss::*verbose* nil) :success)
   (:otherwise :failure)
   )
  )

;;;================================================================================
;;;
;;;                      WHAT-IS CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to obtain and print information about something
;;; corresponding to unformatted data

;;;(defParameter *empty-words* 
;;;  '("the" "a" "an" "of" "from" "to" "those" "these" "that" "s" "?"
;;;    "le" "la" "les" "de" "du" "des" #+MCL "à" #+MICROSOFT-32 "à" "aux" 
;;;    "ce" "ces"))

;;;================================================================================

(defsubdialog 
    _what-is-conversation
    (:label "Get info dialog")
  (:explanation "Master is trying to get info about some data.")
  (:states _wai-entry-state
           _wai-ask-staff
           _wai-get-info
           _wai-process-staff-answer
           _wai-sorry)
  )

;;;------------------------------------------------------------- (WAI) ENTRY-STATE

(defstate 
    _wai-entry-state
    (:label "What is conversation")
  (:entry-state _what-is-conversation)
  (:explanation "Assistant is trying to get object info from its knowledge base.")
  (:transitions
   ;;keep only relevant data
   (:patterns ((* "qu" "est" "ce" "que" ?y "?")
               (?y "c" "est" "quoi" "?" *)
               (* "c" "est" "quoi" ?y "?")
               (?y)  ; when entering from print-concept-documentation
               )
              :keep ?y
              :target _wai-get-info)
   (:otherwise :target _wai-sorry)
   )
  )

;;;--------------------------------------------------------------- (WAI) ASK-STAFF
;;; changing strategy to :collect-answers
;;; rather than sending a broadcast right away, maybe we should
;;; 1. look into local KB
;;; 2. look into our agenda (sa-address)
;;; 3. ask BIBLIO
;;; This would avoid strange answers resulting from misinterpretation of the
;;; query.

(defstate 
    _wai-ask-staff
    (:label "Ask staff.")
  (:explanation "Ask the staff for help.")
  (:send-message :to :all :self PA_ARISTIDE :action :what-is :timeout 2
                 :args `(((:data . ,(read-fact moss::conversation :input))
                          (:language . :fr)
                          ))
                 :strategy :collect-answers
                 )
  (:transitions 
   (:always :target _wai-process-staff-answer))
  )

;;;---------------------------------------------------------------- (WAI) GET-INFO

(defstate 
    _wai-get-info
    (:label "Looking into local KB info.")
  (:explanation "Look into own KB to get info.")
  (:answer-analysis)
  )

(defownmethod =answer-analysis _wai-get-info (conversation input)
  "we try first to get info locally and if it fails ask all agents.
Arguments:
   agent: assistant."
  (declare (special *empty-words*))
  (let* (entry-list)
    (catch :return
           ;; clean input of parasitic words
           (setq input (mapcan #'(lambda (xx) 
                                   (if (member xx *empty-words* :test #'string-equal)
                                       nil
                                     (list xx)))
                         input))
           (moss::vformat "ARISTIDE: _wai-get-info /+++ cleaned input:~& ~S" input)
           ;; if nothing left, then go ask staff
           (unless input
             (throw :return `(:transition ,_wai-ask-staff)))
           ;; check whether the resulting words correspond to entry points
           (setq entry-list (moss::find-best-entries input))
           (moss::vformat "ARISTIDE: _wai-get-info /entry list:~&    ~S" entry-list)
           ;; if nothing left go ask staff
           (unless entry-list 
             (throw :return `(:transition ,_wai-ask-staff)))
           ;; if more than one result, ask for help (not yet)
           ;; get documentation or CV
           ;; print it
           (dolist (item entry-list)
             (send item '=get-documentation :lead "   - ")   
             (send conversation '=display-text *answer*)
             )
           ;; transfer to :success    
           (list :success) 
           )
    ))

;;;------------------------------------------------------ (WAI) PROCESS-STAFF-ANSWER

(defstate 
    _wai-process-staff-answer
    (:label "Process staff answer.")
  (:explanation "Process anwers from the staff agents.")
  (:answer-analysis)
  )

(defownmethod =answer-analysis _wai-process-staff-answer (conversation input)
  "we got an answer from some staff or else the master got impatient.
Arguments:
   agent: assistant."
  (declare (ignore input))
  (let* ((agent (car (has-moss-agent conversation)))
         (answer-list (omas::answer agent))
         (*language* :fr)
         )
    ;(format *debug-io* "~&+++ =resume; HAS-TO-DO: ~S" (has-to-do conversation))
    ;; clean up HAS-TO-DO (contains "answer-there")
    ;(setf (has-to-do conversation) nil)
    ;; answer contains either the word "failure" or a list of lists of strings
    ;; (("Jean-Paul Barthes is a professor at UTC ..."
    ;;  "Dominique Barthes-Biesel is a professor at UTC ...")
    ;;  ("Jean-Paul Barthes has NN publications on:  ...")
    ;;  ...)
    ;; 
    (moss::vformat "~&>>> WAIdialog /Answer to the send-request:~% ~S" answer-list)
    ;; failure if :failure is returned or :error on timeout
    (cond
     ((or (equal "failure" answer-list) (eql :error answer-list))
      `(:transition ,_wai-sorry))
     ((null answer-list)
      (send conversation '=display-text "- Je ne sais pas...")
      (list :success))
     (t 
      (moss::vformat " _wai-ask-staff /answer-list: ~& ~S" answer-list)
      (send conversation '=display-text "~%C'est:")
      ;; each item of the answer list is like 
      ;;    ("personne" ("nom" "...")("prénom" "..."))
      (dolist (answer answer-list)
        ;; answer could be "failure" rather than a list of strings
        (unless (and (stringp answer) (string-equal answer "failure"))
          (dolist (item answer)
            (send conversation '=display-text 
                  (concatenate 'string "~%" (car item)))
            (send conversation '=display-text
                  (format nil ": ~{~%     ~{~A~^, ~}~}" (cdr item))))))
      (list :success)))
    ))

;;;------------------------------------------------------------------- (WAI) SORRY

(defstate 
    _wai-sorry
    (:label "Failure to get info.")
  (:explanation "We could not find anything.")
  (:reset)
  (:text "~%- Désolé, je n'ai aucune information correspondant à ces données.~%")
  (:transitions 
   (:failure))
  )

;;;================================================================================
;;;
;;;                    WHO IS (GET PERSON INFO) CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to print information about a person.

(defsubdialog 
    _get-person-info-conversation
    (:label "Get person info dialog")
  (:explanation "Master is trying to obtain a telephone number.")
  (:states _gpi-entry-state
           _gpi-ask-staff
           _gpi-get-info
           _gpi-sorry)
  )

;;;================================================================================
;;;                   WHO IS CONVERSATION SERVICE FUNCTIONS
;;;================================================================================

;;;------------------------------------------------------------ CHECK-ENTRY-POINTS

(defUn check-entry-points (input)
  "tries to get entry points by combining words.
Arguments:
   input: list of one or more words
Return:
   nil or a list like 
     (\"barthes\" ($T-PERSON-NAME.OF . $E-PERSON.1) ($T-PERSON-NAME.OF . $E-PERSON.2))"
  (let (result entry entry-list expr-list)
    ;; combine all possible words
    (setq expr-list (make-word-combinations input))
    ;; try all possibilities
    (dolist (expr expr-list)
      ;; entry point ?
      (when (setq entry (moss::%is-entry? expr))
        ;; extract corresponding objects
        (setq result (moss::%%get-objects-from-entry-point entry))
        ;; return to the executing environment
        (moss::vformat "...trying ~S in package ~S as entry...~&   => ~S" 
                       expr *package* result)
        (if result (setq entry-list (append entry-list result)))
        ))
    ;; clean list, eliminating tasks whose performative is different of the required
    ;; one
    (setq entry-list (delete-duplicates entry-list :test #'equal))
    (moss::vformat "check-entry-points /Resulting entry list ~&  => ~S" entry-list)
    (mapcar #'cdr entry-list)))

#|
? (albert::check-entry-points '("telephone" "barthes" ))
;***** ...trying "telephone" in package #<Package "SA-ADDRESS"> as entry...
   => ((MOSS::$ENAM.OF . $E-PHONE))
;***** ...trying "barthes" in package #<Package "SA-ADDRESS"> as entry...
   => (($T-PERSON-NAME.OF . $E-PERSON.2) ($T-PERSON-NAME.OF . $E-PERSON.3)
       ($T-PERSON-NAME.OF . $E-PERSON.4))
;***** check-entry-points /Resulting entry list 
  => ((MOSS::$ENAM.OF . $E-PHONE) ($T-PERSON-NAME.OF . $E-PERSON.2)
      ($T-PERSON-NAME.OF . $E-PERSON.3) ($T-PERSON-NAME.OF . $E-PERSON.4))
($E-PHONE $E-PERSON.2 $E-PERSON.3 $E-PERSON.4)
|#
;;;-------------------------------------------------------- MAKE-WORD-COMBINATIONS

(defUn make-word-combinations (word-list)
  ;; when empty list or only one word return
  (unless (cdr word-list) (return-from make-word-combinations word-list))
  ;; otherwise build word combinations
  (let ((input-length (length word-list)) ; e.g. 5
        result)
    (dotimes (nn (1+ input-length)) ; e.g. nn: 0 -> 4
      ;; build cursor to move through the list
      (dotimes (jj nn) ; e.g. jj: 0, then jj: 0->1
        (push
         (apply #'moss::%make-phrase 
                (subseq word-list jj (+ (- input-length nn) jj 1)))
         result)))
    (reverse result)))

#|
? (make-word-combinations '("a" "b" "c" "d"))
("a b c d" "a b c" "b c d" "a b" "b c" "c d" "a" "b" "c" "d")
? (moss::%make-name-for-entry "a b c")
A-B-C
|#
;;;================================================================================
;;;                           WHO IS CONVERSATION STATES
;;;================================================================================

;;;------------------------------------------------------------- (WOI) ENTRY-STATE

(defstate 
    _gpi-entry-state
    (:label "Get person info conversation")
  (:entry-state _get-person-info-conversation)
  (:explanation "Assistant is trying to get the info from its knowledge base.")
  (:transitions
   ;;keep only relevant data
   (:patterns ((* "qui" "est" ?y "?")
               (* "qui" "est" ?y)
               (?y "c" "est" "qui" "?")
               (* "c" "est" "qui" ?y "?")
               )
              :keep ?y
              :target _gpi-get-info)
   (:otherwise :target _gpi-sorry)
   )
  )

;;;--------------------------------------------------------------- (WOI) ASK-STAFF
;;; changing strategy to :collect-answers
;;; rather than sending a broadcast right away, maybe we should
;;; 1. look into local KB
;;; 2. look into our agenda (ADDRESS)
;;; 3. ask BIBLIO
;;; This would avoid strange answers resulting from misinterpretation of the
;;; query.

(defstate 
    _gpi-ask-staff
    (:label "Ask other agents for help.")
  (:explanation "Ask the staff for help.")
  (:send-message :to :all :self PA_ARISTIDE :action :who-is :timeout 4
                 :args `(((:data . ,(read-fact moss::conversation :input))
                          (:language . :fr)
                          ))
                 :strategy :collect-answers
                 )
  (:transitions
   (:on-failure :target _gpi-sorry)
   (:otherwise
    :exec (dolist (answer (read-fact moss::conversation :answer))
            ;; answer could be "failure" rather than a list of strings
            (unless (or (and (stringp answer) (string-equal answer "failure"))
                        (null answer))
              (dolist (item answer)
                (send moss::conversation '=display-text item))))
    :success))
  )

;;;---------------------------------------------------------------- (WOI) GET-INFO

(defstate 
    _gpi-get-info
    (:label "Looking into local KB info.")
  (:explanation "Look into own KB to get answer to who-is?.")
  (:answer-analysis)
  )

(defownmethod =answer-analysis _gpi-get-info (conversation input)
  "we got an answer from some staff or else the master got impatient.
Arguments:
   conversation: the conversation object
   input: a copy of the FACTS/input area"
  (let* (entry-list)
    (catch :return
           ;; clean input of parasitic words
           (setq input (mapcan #'(lambda (xx) 
                                   (if (member xx *empty-words* :test #'string-equal)
                                       nil
                                     (list xx))
                                   )
                         input))
           (moss::vformat "ARISTIDE: _gpi-get-info /+++ cleaned input:~& ~S" input)
           ;; if nothing left, then ask staff
           (unless input
             (throw :return `(:transition ,_gpi-ask-staff)))
           ; check whether the resulting words correspond to entry points
           (setq entry-list (check-entry-points input))
           (moss::vformat "ARISTIDE: _gpi-get-info /entry list:~& ~S" entry-list)
           ;; if nothing left go ask staff
           (unless entry-list 
             (throw :return `(:transition ,_gpi-ask-staff)))
           ;; if more than one person ask for help (not yet)
           ;; get documentation or CV
           ;; print it
           (dolist (item entry-list)
             (send item '=get-documentation)
             (send conversation '=display-text *answer*))
           ;; transfer to :success    
           (list :success) 
           )
    ))

;;;------------------------------------------------------------------- (WOI) SORRY

(defstate 
    _gpi-sorry
    (:label "Sorry: no info.")
  (:explanation "We could not find anything.")
  (:reset)
  (:text ("~%- Désolé, je n'en ai pas la moindre idée.~%"
          "- Je ne sais pas.~%"))
  (:transitions 
   (:failure))
  )

;;;================================================================================
;;;================================================================================
;;;
;;;                  Conversations and Sub-Conversations
;;;
;;;================================================================================

;;; we declare the top-level conversations
;;; the entry point of the dialog is _MAIN-CONVERSATION

(eval-when (:compile-toplevel :load-toplevel :execute)
  (proclaim '(special 
              _MAIN-CONVERSATION
              _PROCESS-CONVERSATION
              )))

;;;================================================================================
;;;
;;;                       MAIN CONVERSATION (CONTROL LOOP)
;;;
;;;================================================================================

;;; create a dialog header to be used for the main conversation

(defindividual 
    MOSS-DIALOG-HEADER
    (HAS-MOSS-LABEL "Main conversation loop")
  (HAS-MOSS-EXPLANATION "This is the main conversation loop.")
  (:var _main-conversation))

;; declare that conversation, replacing default conversation in the agent structure
(setf (omas::dialog-header PA_ARISTIDE) _main-conversation)


;;;===== states are declared as global variables
;;; the set of states can be considered as a plan to be executed for conducting the
;;; conversation with the master


;;;--------------------------------------------------------------------------------
(eval-when (:compile-toplevel :load-toplevel :execute)
  (proclaim '(special 
              _mc-entry-state
              _mc-get-input
              _mc-more?
              _mc-process
              _mc-sleep)))
;;;--------------------------------------------------------------------------------

;;;--------------------------------------------------------------- (MC) ENTRY-STATE
;;; ENTRY STATE
;(defun ff () (start-conversation _main-conversation :first-time t))
;(defun gg () (start-conversation _main-conversation))

;;; define an entry instance and record it in a global variable.
;;; ****** we should enter a frame on the conversation frame list, to handle
;;; success or failure returns from sub-dialogs. I.e. we should make a transition
;;; on the success or failure state of the main conversation.
;;; entry (_mc-failure _mc-success _mc-entry-state nil) ; no goal 

;;; Action:
;;;   =EXECUTE
;;;   - get ACTION if any from the dialog header
;;;   - reset conversation: HAS-MOSS-FACTS, HAS-MOSS-TASK, HAS-MOSS-TASK-LIST
;;;   - if ACTION IS non NIL, set conversation HAS-GOAL with a pattern obtaines from
;;;        sending ACTION a =create-pattern message
;;;     otherwise, reset conversation HAS-GOAL
;;;   - display welcome message
;;;
;;;   <Master enters text>
;;;
;;;   =RESUME
;;;   - if INPUT contains one of the abort commands, throws to :DIALOG-ERROR
;;;   - makes a transition to _MC-PROCESS

(defstate _mc-entry-state
    (:entry-state _main-conversation)
  (:label "Début du dialogue")
  (:explanation 
   "Initial state when the assistant starts a conversation. Send a welcome message ~
    and wait for data. Also entered on a restart following an abort.")
  (:reset-conversation)
  (:text "Attention! Ce dialogue est trés limité. Chaque phrase est indépendante ~
          et l'on ne peut utiliser des pronoms référents.~%")
  (:question-no-erase
   ("- Bonjour ! que puis-je faire pour vous ?"
    "- Salut ! Que voudriez-vous savoir ?" 
    "- Bonjour ! Je vais faire de mon mieux pour répondre à vos ~
     questions. Mais, rappelez-vous que mon QI est faible."))
  (:transitions 
   (:always :target _mc-process))
  )

;;;----------------------------------------------------------------- (MC) GET-INPUT

(defstate _mc-get-input
    (:label "Get input")
  (:explanation
   "State in which the user inputs his request.")
  (:reset)
  (:question "- Je vous écoute....")
  (:transitions
   (:always :target _mc-process)))

;;;-------------------------------------------------------------------- (MC) MORE?
;;; Action:
;;;   - If answer is "rien", "non" transition to SLEEP
;;;   - if answer is "oui" transition to GET-INPUT
;;;   - anything else sends to PROCESS

(defstate _mc-more?
    (:label "More?")
  (:explanation "Asking the user it he wants to do more interaction.")
  (:reset-conversation)
  (:question-no-erase
   ("~%- Que puis-je faire d'autre pour vous ?"
    "~%- Y a-t-il autre chose que je puisse faire pour vous ?"
    "~%- Avez-vous d'autres questions ?"
    "~%- OK.")
   ) 
  (:transitions 
   (:starts-with ("rien") :target _mc-sleep)
   (:no :target _mc-sleep)
   (:yes :target _mc-get-input)
   (:otherwise :target _mc-process)))

;;;------------------------------------------------------------------- (MC) PROCESS

(defstate _mc-process
    (:label "Process state of main conversation")
  (:explanation
   "Here we call the main process organized as a sub-conversation.")
  (:transitions
   (:always :sub-dialog _process-conversation
            :success _mc-more?
            :failure _mc-more?)))

;;;--------------------------------------------------------------------- (MC) SLEEP

(defstate _mc-SLEEP
    (:label "Nothing to do")
  (:explanation
   "User said she wanted nothing more.")
  (:reset)
  (:text "- OK. J'attends que vous soyez prêt.~%")
  (:question-no-erase 
   "- Réveillez moi en tapant ou en disant quelque chose...")
  (:transitions
   (:always :target _mc-process)))

;;;================================================================================
;;;
;;;                      MAIN CONVERSATION (EXECUTION PART)
;;;
;;;================================================================================

;;; this conversation is intended to process the input from te user by first
;;; determining the performative, then the list of possible tasks

(defsubdialog 
    _process-conversation 
    (:label "Process conversation")
  (:explanation 
   "Processing steps of the main conversation.")
  (:states _mc-eliza
           _mc-find-performative
           _mc-entry-state
           _mc-failure
           _mc-find-performative
           _mc-find-task
           _mc-select-task
           _mc-task-dialog)
  )

;;;===============================================================================
;;;                            Service functions
;;;===============================================================================

;;;-------------------------------------------------------------------- PRINT-TIME

(defUn print-time (conversation &optional (language *language*))
  (multiple-value-bind (second minute hour date month year)
      (decode-universal-time (get-universal-time))
    (send conversation '=display-text
          (case language
            (:en
             (format nil "Current time is ~S:~S:~S and today is the ~S/~S/~S~%"
               hour minute second date month year))
            (:fr
             (format nil 
                 "Il est actuellement ~S:~S:~S et nous sommes aujourd'hui le ~
                      ~S/~S/~S~%"
               hour minute second date month year)))
          )))

;;;===============================================================================
;;;                            Conversation states
;;;===============================================================================

;;;--------------------------------------------------------- (MC) FIND-PERFORMATIVE

(defstate _mc-find-performative
    (:entry-state _process-conversation)
  (:process-state _main-conversation)
  (:label "Find performative")
  (:explanation 
   "Process what the user said trying to determine the type of performative ~
    among :request :assert :command. Put the result if any into the performative ~
    slot of the conversation object.")
  (:transitions 
   ;; time?
   (:patterns (("Quelle" "heure" *))
              :exec (print-time moss::conversation :fr) :success)
   ;; now really select performatives
   (:patterns (("qui" *)
               ("quel" "est" *)
               ("quelle" "est" *)
               ("quels" "sont" *)
               ("quelles" "sont" *)
               ("qu" "est-ce" "que" *)
               ("quoi" *)
               ("quand" *)
               ("où" *)
               ("pourquoi" *)
               ("qui" *)
               ("combien" *)
               ("comment" *)
               ("est-ce" "que" *)
               ("est-il" *)
               ("est" "il" *)    ; interface vocal ?
               ("est-elle" *)
               ("est" "elle" *)
               ("sont-ils" *)
               ("sont" "ils" *)
               ("sont-elles" *)
               ("sont" "elles" *)
               ("a-t-il" *)
               ("a-t-elle" *)
               ("ont-ils" *)
               ("ont-elles" *)
               (* "c" "est" "quoi" *)
               (* "c" "est" "qui" *)
               (* "?")
               )
              :set-performative (list :request)
              :target _mc-find-task)
   (:patterns ((* "notez" *)
               (* "noter" *))
              :set-performative (list :assert)
              :target _mc-find-task)
   (:otherwise 
    :set-performative (list :command)
    :target _mc-find-task))
  )

;;;--------------------------------------------------------------------- (MC) ELIZA

(moss::defstate _mc-ELIZA
    (:label "ELIZA")
  (:explanation
   "Whenever MOSS cannot interpret what the user is saying, ELIZA is called to ~
    do some meaningless conversation to keep the user happy. It then record the ~
    master's input and transfers to find performative state.")
  (:eliza)
  (:transitions
   (:always :target _mc-find-performative))
  )

;;;------------------------------------------------------------------- (MC) FAILURE

(defstate _mc-failure
    (:label "Failure state of main conversation")
  (:explanation
   "Failure state is entered when we return from a sub-dialog with a :failure tag. ~
    We check for more tasks to perform (listed in the task-list slot of the ~
    conversation object). If there are more, we execute the first one. If there ~
    are no more, we return with a failure tag.")
  (:answer-analysis)
  )

(defownmethod
    =answer-analysis _mc-failure (conversation input)
  (declare (ignore input))
  (let* ((task-list (HAS-MOSS-TASK-LIST conversation))
         task)
    (if task-list
        (progn
          ;; record next task
          (setq task (pop task-list))
          ;; remove it from task-list
          (send conversation '=replace 'HAS-MOSS-TASK-LIST task-list)
          ;; add it to conversation
          (send conversation '=replace 'HAS-MOSS-TASK (list task))
          ;; transfer to sub-dialog
          `(:transition ,_mc-task-dialog))
      ;; when no more tasks we return :failure and the state variable still contains
      ;; :failure. The crawler will return one more level until hitting a non
      ;; failure return containing a specific state (e.g. the one for the main
      ;; conversation)
      (list :failure)
      )))

;;;----------------------------------------------------------------- (MC) FIND-TASK

(defstate _mc-find-task
    (:label "Find task")
  (:explanation 
   "Process the input to determine the task to be undertaken. Combines the words ~
    from the sentence to see if they are entry points for the index property of ~
    any task. Collect all tasks for which there is an index in the sentence.")
  (:answer-analysis)
  )

(defownmethod
    =answer-analysis _mc-find-task (conversation input)
  "Using input to find tasks If none failure, if one, OK, if more, must ask user to ~
   select one. Checks for an entry point for an index of a task.
Arguments:
   conversation: current conversation
   input: list containing the input as a list of words"
  (let* ((performative (read-fact conversation :PERFORMATIVE))
         task-list)
    (moss::vformat "_mc-find-task /performative: ~S, package: ~S, input:~%  ~S" 
                   performative *package* input)
    ;; try to find a task from the input text
    (setq task-list (moss::find-objects 
                     '(TASK 
                       (HAS-INDEX-PATTERN 
                        (TASK-INDEX (HAS-INDEX :is :?))))
                     input :all-objects t)) 
    (moss::vformat "_mc-find-task /possible tasks:~%  ~S" task-list)
    ;; filter tasks that do not have the right performative
    (setq task-list
          (mapcan #'(lambda (xx) 
                      (if (intersection performative (HAS-PERFORMATIVE xx))
                          (list xx)))
            task-list))
    (moss::vformat "_mc-find-task /task list after performative check: ~S" task-list)
    (cond 
     ;; if empty, ask ELIZA
     ((null task-list)
      `(:transition ,_mc-ELIZA))
     ;; if one, then OK
     ((null (cdr task-list))
      ;; save results, make the task the conversation task
      (setf (HAS-MOSS-TASK conversation) task-list)
      `(:transition ,_mc-task-dialog))
     (t 
      ;; otherwise must select one from the results
      (setf (HAS-MOSS-TASK-LIST conversation) task-list)
      `(:transition ,_mc-select-task)))
    ))

;;;--------------------------------------------------------------- (MC) SELECT-TASK

(defstate 
    _mc-select-task
    (:label "mc select task")
  (:explanation "we have located more than one task. We rank the tasks by computing ~
                 the average weight of the terms in the task index slot. We then ~
                 record the list of tasks into the statte-context task-list slot ~
                 and activate the highest ranking task.")
  (:answer-analysis) 
  )

(defownmethod =answer-analysis _mc-select-task (conversation input)
  "we have located more than one task. We rank the tasks by computing ~
   the MYCIN combination weight of the terms in the task index slot. We then ~
   record the list of tasks into the conversation task-list slot ~
   and activate the highest ranking task."
  ;(declare (ignore input))
  ;; the list of tasks is in the HAS-MOSS-TASK-LIST slot. Tasks are defined in the application
  ;; package
  (let* ((task-list (HAS-MOSS-TASK-LIST conversation))
         patterns weights result pair-list selected-task level word-weight-list)
    (moss::vformat "_select-task /input: ~S~&  task-list: ~S" input task-list)
    ;; first compute a list of patterns (combinations of words) from the input
    (setq patterns (mapcar #'car (moss::generate-access-patterns input)))
    (moss::vformat "_select-task /input: ~S~&  generated patterns:~&~S" 
                   input patterns)
    ;; then, for each task
    (dolist (task task-list)
      (setq level 0)
      ;; get the weight list
      ;(setq weights (HAS-INDEX-WEIGHTS task))
      (setq weights (moss::%get-INDEX-WEIGHTS task))
      (moss::vformat "_select-task /task: ~S~&  weights: ~S" task weights)
      ;; check the patterns according to the weight list
      (setq word-weight-list (moss::%get-relevant-weights weights patterns))  
      (moss::vformat "_select-task /word-weight-list:~&  ~S" word-weight-list)
      ;; combine the weights
      (dolist (item word-weight-list)
        (setq level (+ level (cadr item) (- (* level (cadr item))))))
      (moss::vformat "_select-task /level: ~S" level)
      ;; push the task and weight onto the result list
      (push (list task level) result)
      )
    (moss::vformat "_select-task /result:~&~S" result)
    ;; order the list
    (setq pair-list (sort result #'> :key #'cadr)) 
    (moss::vformat "_select-task /pair-list:~&  ~S" pair-list)
    ;; keep the first task whatever its score
    (setq selected-task (caar pair-list))
    ;; remove the task that have a weight less than task-threshold (default 0.4)
    (setq pair-list
          (remove nil 
                  (mapcar 
                      #'(lambda (xx) 
                          (if (>= (cadr xx) (omas::task-threshold omas::*omas*)) xx))
                    pair-list)))
    ;; if task-list is empty then return th? first saved task
    ;; this may not be a good policy if the score is too low
    (if (null pair-list)
        (progn
          ;; reset the task-list slot of the conversation object
          (setf (HAS-MOSS-TASK-LIST conversation) nil)
          ;; put the saved task into the task slot
          (setf (HAS-MOSS-TASK conversation) (list selected-task))
          ;; go to task-dialog
          `(:transition ,_mc-task-dialog)
          )
      (progn
        ;; remove the weights
        (setq task-list (mapcar #'car pair-list))
        ;; select the first task of the list
        (setq selected-task (pop task-list))
        (moss::vformat "_select-task /selected task: ~S" selected-task)
        ;; save the popped list in the task-list slot of the conversation object
        (setf (HAS-MOSS-TASK-LIST conversation) task-list)
        (setf (HAS-MOSS-TASK conversation) (list selected-task))
        ;; go to task-dialog
        `(:transition ,_mc-task-dialog)))
    ))

;;;--------------------------------------------------------------- (MC) TASK-DIALOG

(defstate _mc-task-dialog
    (:label "Task dialog")
  (:explanation 
   "We found a task to execute (in the GOAL slot of the conversation). We activate ~
    the dialog associated with this task.")
  ;; we launch the task dialog as a sub-conversation
  ;; the task contains the name of a sub-conversation header, e.g. _get-tel-nb
  (:answer-analysis)
  )

;;;********** must check for possible problems
(defownmethod
    =answer-analysis _mc-task-dialog (conversation input)
  "We prepare the set up to launch the task dialog.
Arguments:
   conversation: current conversation"
  (declare (ignore input))
  (let* ((task-id (car (send conversation '=get 'HAS-MOSS-TASK))))
    (moss::vformat "_mc-task-dialog /task-id: ~S dialog: ~S" 
                   task-id (HAS-DIALOG task-id))
    ;; we launch the task dialog as a sub-conversation
    ;; the task contains the name of a sub-conversation header, e.g. _get-tel-nb
    `(:sub-dialog ,(car (HAS-DIALOG task-id)) :failure ,_mc-failure)
    ))

;;;================================================================================
;;;
;;;                             ELIZA CONVERSATION
;;;
;;;================================================================================

;;; this conversation is intended to do small talk.


(defParameter *eliza-rules*
  '((("hello")
     "Bonjour ! Je m'appelle ARISTIDE."
     "Hello there?")
    (("Bom" "dia")
     "Desculpe ! Não falo Portugues."
     "Bom dia. Por favor fale em frances.")
    (("Salut" (?* ?x))
     "Salut ! Je m'appelle ARISTIDE."
     "Salut ?")
    (("bonjour")
     "Bonjour ! Je m'appelle ARISTIDE.")
    (("bonjour" (?* ?x))
     "Bonjour, Boss !")
    (("hi")
     "Bonjour ! Je m'appelle ARISTIDE. Mon anglais n'est pas bon.")     
    (("comment" "allez-vous" (?* ?x))
     "Très bien merci."
     "Vraiment, ma santé vous intéresse ? J'en suis trés flatté.")
    (("qui"  * "Barthès")
     "Jean-Paul Barthès est professeur d'intelligence artificielle à ~
      l'Université de Technologie de Compiègne.")
    (("qui"  * "Barthes")
     "Jean-Paul Barthès est professeur d'intelligence artificielle à ~
      l'Université de Technologie de Compiègne.
      Notez toutefois que Barthès s'écrit avec un è.")
    (("laissez" "tomber")
     "OK. Pas de problème.")
    (("merci" *)
     "Avec plaisir.")
    (("tout" *)
     "Vraiment ?"
     "Quelle exigence !")
    ((* "machine" *)
     "Est-ce que les machines vous dérangent ?"
     "Que pensez-vous des machines ?"
     "Pourquoi mentionnez-vous les machines ?"
     "Les machines ont-elles un rapport avec votre problème ?")
    ((* "stupide" *)
     "S'il vous plaît, restez correct !"
     "Je suis peiné par votre remarque.")
    ((?x "ne" "marche" "pas" *)
     "Bien sûr que ?x marche !"
     "Pourquoi dites-vous que ?x ne marche pas ?")
    (("à" "l" "aide")
     "Vous voulez de l'aide à quel sujet?")
    (("help")
     "Vous voulez de l'aide à quel sujet?")
    ((* "désolé" *)
     "S'il vous plaît, ne vous excusez pas."
     "Des excuses ne sont pas nécessaires"
     "Que ressentez-vous quand vous vous excusez ?")
    ((* "Je" "me" "souviens" ?y)
     "Vous souvenez-vous souvent ?y ?"
     "Vous souvenir ?y vous rappelle-t-il autre chose ?"
     "De quoi d'autre vous souvenez-vous ?"
     "Pourquoi vous souvenez-vous ?y maintenant ?"
     "Qu'est-ce qui dans la situation présente vous fait vous souvenir ?y ?"
     "Quel est le lien entre moi et ?y ?")
    ((* "vous" "rappelez-vous" ?y)
     "Pensez-vous que j'aurais oublié ?y ?"
     "Pourquoi pensez-vous que je devrais me rappeler ?y maintenant ?"
     "Qu'y a-t-il à propos de ?y ?"
     "Vous avez mentionné ?y ?")
    ((* "si" ?y)
     "Pensez-vous que ?y soit probable ?"
     "Souhaitez-vous ?y ?"
     "Que pensez-vous de ?y ?"
     "Vraiment -- si ?y ?")
    (("que" * "faire" "pour" "moi")
     "Tant de choses..."
     "A ce moment précis, je ne vois pas."
     "Qu'est-ce qui vous ferait plaisir ?")
    ((* "J" "ai" "rêvé" ?y)
     "Vraiment -- ?y"
     "Avez-vous déjà rêvé ?y tout eveillé ?"
     "Avez-vous déjà ?y auparavant ?")
    ((* "rêve" *)
     "Qu'est-ce que ce rêve vous suggère ?"
     "Rêvez-vous souvent ?"
     "Quelles personnes apparaissent dans vos rêves ?"
     "Ne pênsez-vous pas que les rêves sont liés à votre problème ?")
    ((* "ma" "mère" ?y)
     "Qui d'autre dans votre famille ?y ?"
     "Parlez moi de votre famille.")
    ((* "mon" "père" *)
     "Votre père ?"
     "A-t-il une grande influence sur vous ?"
     "Qu'est-ce qui vous vient à l'esprit quand vous pensez à votre père ?")
    ((* "Je" "voudrais" ?y)
     "Que signifierait ?y ?"
     "Pourquoi voudriez-vous ?y ?")
    ((* "Je" "suis" "content" ?y)
     "Vous ai-je aidé à être content ?y ?"
     "Qu'est-ce qui vous rend heureux en ce moment ?"
     "Pouvez-vous m'expliquer pourquoi vous êtes tout à coup content ?y ?")
    ((* "Je" "déprime" *)
     "J'en suis fort désolé."
     "je suis sûr que ce n'est pas drïle d'être déprimé.")
    ((?x "est" "comme" ?y)
     "Quelle resemblance voyez-vous entre ?x et ?y ?"
     "En quel sens ?x est comme ?y ?"
     "Pensez-vous qu'il y ait vraiment un lien ?"
     )
    ((* "pareils" *)
     "Comment ça ?"
     "Quelle resemblance y a-t-il ?")
    ((* "pareil" *)
     "Quel autre lien voyez-vous ?")
    
    ((* "J" "étais" ?y)
     "L'étiez-vous vraiment ?"
     "Je me doutais que vous étiez ?y"
     "Pourquoi me dites-vous que vous étiez ?y maintenant ?")
    ((* "Je" "suis" ?y)
     "En quel sens êtes-vous ?y ?"
     "Souhaitez vous être ?y ?")
    ((* "suis-je" ?y)
     "Croyez-vous être ?y ?"
     "Désirez-vous être ?y ?"
     "Qu'est-ce que cela ferait si vous étiez ?y ?")
    ((* "suis" *)
     "Pourquoi dites-vous \"SUIS\" ?"
     "Je ne comprends pas cela.")
    ((* "êtes-vous" ?y)
     "Pourquoi êtes-vous intéressé par savoir si je suis ?y ou pas ?"
     "Préfèreriez-vous que je ne sois pas ?y ?")
    ((* "vous" "êtes" ?y)
     "Qu'est-ce qui vous fait penser que je suis ?y ?")
    ((* "parce" "que" *)
     "Est-ce la seule raison ?"
     "Quelle autre raison pourrait-il y avoir ?"
     "Est-ce que cette raison peut expliquer autre chose ?")
    ((* "étiez-vous" ?y)
     "Peut-être étais-je ?y"
     "Qu'en pensez-vous ?"
     "Et si j'avais été ?y ?")
    ((* "je" "ne" "peux" "pas" ?y)
     "Peut-être pouvez-vous ?y maintenant ?"
     "Et si vous pouviez ?y ?")
    ((* "Je" "ressents" ?y)
     "Ressentez-vous ?y souvent ?")
    ((* "J" "ai" "ressenti" *)
     "Qu'avez-vous ressenti d'autre ?")
    ((* "pourquoi" "ne" "pas" ?y)
     "Ne devriez-vous pas ?y vous-même ?"
     "Pensez-vous que je pourrais ?y ?"
     "Peut-petre pourrais-je?y en temps utile")   
    ((* "oui" *)
     "Vous me semblez très positif."
     "Vos êtes sûr ?"
     "Je comprends.")
    ((* "non" *)
     "Pourquoi non ?"
     "Vous me semblez très négatif."
     "Est-ce que vous dites \"NON\" simplement pour être négatif ?")
    ((* "quelqu" "un" *)
     "Pouvez-vous être plus précis ?")
    ((* "tout" "le" "monde" *)
     "Sûrement pas tout le monde."
     "Pensez-vous à quelqu'un en particulier ?"
     "Qui par exemple ?"
     "Vous pensez à quelqu'un de précis ?")
    ((* "toujours" *)
     "Pouvez-vous donner un exemple précis ?"
     "Quand ?"
     "Toujours -- vraiment ?")
    ((* "peut-être" *)
     "Vous n'avez pas l'air tout à fait sûr.")
    ((* "est" ?y)
     "Vous pensez qu'il ne pourrait pas être ?y ?"
     "Il est peut-être ?y")
    ((* "OK" *)
     "Bon, comme vous voudrez.")
    ((* "zouave" *)
     "Bachibouzouk."
     "Moule à gaufre"
     "Restez poli, tonnerre de Brest!")
    ((* "connard" *)
     "Je me plaindrai à mon syndicat..."
     "C'est ça, c'est ça...")
    ((* "imbécile" *)
     "Oh!"
     "Crâne d'oeuf!"
     "Crétin des Alpes!")
    ((* "idiot" *)
     "Pas tant que ça..."
     "Pas plus que la moyenne."
     "Si vous le dites...")
    
    
    ((*)
     "Je ne vois pas ce que vous voulez exactement."
     "Je ne suis pas sûr de comprendre ce que vous voulez."
     "Pouvez-vous reformuler votre question s'il vous plaît ?"
     "Désolé, je n'ai pas compris."
     "Aïe ! Je n'ai pas compris."
     "Pouvez-vous dire cela différemment ?"
     "Pardon ?"
     "Quoi ?"
     "Comment ?"
     "J'ai entendu ?x , mais je ne comprends pas."
     )
    ))

(format t "~%;*** MOSS v~A - ARISTIDE dialogs loaded ***" moss::*moss-version-number*)

;;;=============================================================================== 

:EOF
