;;;-*- Mode: Lisp; Package: "PATIENT" -*-
;;;===============================================================================
;;;11/05/09
;;;             AGENT PATIENT: DIALOG (file PATIENT-dialog.lisp)
;;;             Copyright Barthes@UTC, 2011
;;;
;;;=============================================================================== 

;;; Note that this file must be in UTF-8 format (like the other agent files)
;;; The top-level dialog works in English, French, Portuguese (Brasilian), or Japanese
;;; For any other language the questions and answers must be translated and added
;;; to the already existing ones.

#|
PATIENT is a therapist assistant agent. Its has a single main dialog.
 
2011
 0423 creation
 0509 adding a function to synthesize states from the ontology
|#

(in-package :PATIENT)

;; this is used by the Japanese language, can be removed if not needed
(eval-when (:load-toplevel :compile-toplevel :execute)
  (push '?? moss::*yes-answers*)
  (push 'OK moss::*yes-answers*)
  (push '??? moss::*no-answers*)
  )
  
;;;================================================================================
;;;
;;;                         PATIENT macros.
;;;
;;;================================================================================

(defmacro clean-topics ()
  `(progn
     (post-topics pa_patient (omas::com-window pa_patient) 
       '("" "" "" "" "" "" "" ""))
	  (clean-topic-info)))
	   
(defmacro show-topics (topic-list)
  `(post-topics pa_patient (omas::com-window pa_patient) ',topic-list))
  
;;;================================================================================
;;;
;;;                         PATIENT functions
;;;
;;;================================================================================

;;; This function generates states automatically from the ontology

(defun make-state-from-concept (concept-ref &key single-state)
   "takes a concept label (string) and builds the corresponding conversation ~
    _<label>-CONVERSATION, then the dialog entry-state _<label>-POST-TOPICS; ~
	creates symbols for subtopic states _<item>-ENTRY; creates a failure state ~
    _<label>-FAILURE"
   (let ((topic-id (car (access `("topic" ("label" :is ,concept-ref)))))
         name sub-topic-list label result entry-name try-again xx)
	(flet ((make (zz) (if moss::*debug* (print zz) (eval zz))))
	 (when topic-id
	   (setq sub-topic-list 
	     (reduce #'append (mapcar #'has-label (has-sub-topics topic-id))))
	   (setq sub-topic-list (mapcar #'string-upcase sub-topic-list))
	   (setq label (moss::%%make-name-string concept-ref :plain))
	   (setq entry-name (intern (concatenate 'string "_" label "-ENTRY-STATE")))
	   (setq name (intern (concatenate 'string "_" label "-POST-TOPICS")))
	   (setq no-visit (intern (concatenate 'string "_" label "-FAILURE")))
	   (setq try-again (intern (concatenate 'string "_" label "-TRY-AGAIN")))
	   (setq sub-conversation (intern(concatenate 'string "_" label "-CONVERSATION")))

	   ;; we enter a subconversation, must create a subdialog header
	   (make 
	    `(defsubdialog ,sub-conversation
            (:label "Palace conversation")
            (:explanation "We talk about the palace.")
            (:states ,entry-name
                     ,name
					 ,try-again
                     ,no-visit)
		     ))	  
			 
	  (when sub-topic-list
			 
	   ;; create the entry state, posting the welcome message
	   (make
	    `(defstate ,entry-name
	       (:label ,(format nil "~A welcome" concept-ref))
		   (:explanation ,(format nil "Post ~A welcome message." concept-ref))
	       (:entry-state ,sub-conversation)		   
	       (:question ,(has-welcome topic-id))
		   (:transitions
		      (:always :target ,name))
		   ))
		   
       ;; then build state
	   (make
	    `(defstate ,name
	       (:label ,(format nil "Post ~A topics." concept-ref))
		   (:explanation ,(format nil "Topic is ~A." concept-ref))
	       (:execute-preconditions
	         (post-topics pa_patient (omas::com-window pa_patient) ',sub-topic-list))
	       (:question ,(or (has-question topic-id)
		                    "Please select a topic on the menu."))
	       (:transitions 
	         ,@(dolist (item sub-topic-list (reverse result))
		        (push `(:patterns (((?* ?x) ,@(moss::%make-word-list item) (?* ?y)))
				          :sub-dialog 
						    ,(intern (concatenate 'string 
							           "_" 
						               (moss::%%make-name-string item :plain)
								       "-CONVERSATION")))
					  result))
		     (:otherwise :target ,try-again)
		     )))
		
		;; ask user to make a selection
	   (make
	    `(defstate ,try-again
	       (:label ,(format nil "~A try again" concept-ref))
		   (:explanation ,(format nil "In state ~A ask user for a choice." concept-ref))
	       (:question ,(or (has-try-again topic-id)
		                   "Please make a choice by clicking the menu."))
	       (:transitions 
	         ,@(dolist (item sub-topic-list (reverse result))
		        (push `(:patterns (((?* ?x) ,item (?* ?y)))
				          :sub-dialog ,(intern (concatenate 'string "_" item "-CONVERSATION")))
					  result))
		     (:otherwise :target ,no-visit))
		   ))
		   
		;; make failure state
		(make 
		 `(defstate ,no-visit
		    (:label ,(format nil "~A failure state." concept-ref))
		    (:explanation ,(format nil "Topic ~A failed." concept-ref))
		    ,@(if (setq xx (has-failure topic-id)) `((:question  ,xx)))
		    (:transitions
              (:always :exec (clean-topics) :success))))
			  
		;; call sub-conversations recursively
		(unless single-state
		 (dolist (item sub-topic-list)
		   (make-state-from-concept item)))		   			  
	    )
		
	  ;; if there are no subtopics, we create a simple state and return from the
	  ;; previous subdialog
	  (unless sub-topic-list
	   (make 
	    `(defstate ,entry-name
	      (:label ,(format nil "~A state." concept-ref))
		  (:explanation ,(format nil "Topic is ~A." concept-ref))
	      (:entry-state ,sub-conversation)
          (:question ,(or (has-welcome topic-id)
		                  "Sorry, I do not have enough info on this topic."))
	      ,@(if (setq xx (has-welcome topic-id)) `((:question ,xx)))
	      (:transitions 
		    (:always :exec (clean-topics) :success)
		   ))))
	   t)
	   )))

#|
(make-state-from-concept "welcome")
|#
;;;================================================================================
;;;================================================================================
;;;
;;;                       PATIENT CONVERSATION (CONTROL LOOP)
;;;
;;;================================================================================
;;;================================================================================

;;; create a dialog header to be used for the main conversation

(defindividual 
  MOSS-DIALOG-HEADER
  (HAS-MOSS-LABEL "Main conversation loop")
  (HAS-MOSS-EXPLANATION "This is the main conversation loop.")
  (:var _main-conversation))

;; declare that conversation, replacing default conversation in the agent structure
(setf (omas::dialog-header PA_PATIENT) _main-conversation)

;;;------------------------------------------------------------- (DOC) ENTRY-STATE

(defstate
  _doc-entry-state
  (:label "PATIENT dialog entry")
  (:entry-state _main-conversation )
  (:explanation "PATIENT is starting session.")
  ;(:execute-preconditions
  ;    (setf (omas::answer-timeout pa_patient) 10))
  (:question "Hello!")
  (:transitions
    (:always :target _doc-get-name))
  ) 
  
;;;--------------------------------------------------------------------- (DOC) END

(defstate
  _doc-end
  (:label "Closing state")
  (:explanation "PATIENT is ending session.")
  (:question "Nice talking with you. Will do that later another time. Bye...")
  (:transitions
    (:always :success))
  )

;;;---------------------------------------------------------------- (DOC) GET-NAME 

(defstate
  _doc-get-name
  (:label "Get name")
  (:explanation "PATIENT is asking patient's name.")
  (:question "What is your name?")
  (:transitions
    ;; failure case must come before the rules for some reason...
    (:on-failure :exec 
	   (progn (send moss::conversation '=display-text "Nice meeting you."
	                :erase t)
	          (sleep 1.5))
	   :sub-dialog _welcome-conversation :success _doc-end :failure _doc-end)
    (:rules
	   ((((?* ?x) "my" "name" "is" (?* ?y))
	     "Hi, ?y. Nice meeting you.")
	    (((?* ?x) "my" "friends" "call" "me" (?* ?y))
	     "Hi, ?y. Nice meeting you.")
	    (((?* ?x) "call" "me" (?* ?y))
	     "Hi, ?y. Nice meeting you.")
	    (((?* ?y))
		 "Hi, ?y. Nice meeting you."))
		:display-result :erase :sub-dialog _welcome-conversation
		      :success _doc-end :failure _doc-end)
	)
  )

;;;----------------------------------------------------------------- (DOC) NO-VISIT

(defstate
  _doc-no-visit
  (label "No visit")
  (explanation "Patient did not answer or answered negatively.")
  (:question "Ah, have you been,...?")
  (:transitions
     (:always :target _doc-end))
  )
  
;;; The rest of the states is synthesized from the ontology


(make-state-from-concept "WELCOME")

;;;=============================================================================== 

(format t "~%;*** MOSS v~A - PATIENT dialogs loaded ***" moss::*moss-version-number*):EOF